"""

This module contains routines for retrieving galaxy catalogs from various photometric surveys.

"""

import os
import sys
import numpy as np
from astLib import *
import astropy.table as atpy
import astropy.io.fits as pyfits
import astropy.io.ascii as pyascii
import re
import urllib.request, urllib.parse, urllib.error
import urllib.request, urllib.error, urllib.parse
import pylab
import subprocess
import time
import zCluster
import requests 
try:
    from dl import queryClient as qc
except:
    print("WARNING: Failed to import dl module - retrievers that use NOAO Data Lab will not work.")
try:
    import pyvo
except:
    print("WARNING: Failed to import pyvo module - retrievers that use TAP access (e.g. RubinDP0) will not work.")
from astropy.io.votable import parse_single_table 

#-------------------------------------------------------------------------------------------------------------
CACHE_DIR=os.environ['HOME']+os.path.sep+".zCluster"+os.path.sep+"cache"

#-------------------------------------------------------------------------------------------------------------
def makeCacheDir():
    """Makes the cache directory where catalogs (and sometimes other survey-specific data) are stored.
    
    """
    os.makedirs(CACHE_DIR, exist_ok = True)

#------------------------------------------------------------------------------------------------------------
def getRetriever(database, maxMagError = 0.2):
    """Given the name of a survey database, return a function that can retrieve a galaxy catalog, and an 
    associated options dictionary.
    
    Args:
        database (str): Name of the survey database - relatively well-tested options are: SDSSDR12, S82, 
            DESY3, DESDR1, KiDSDR4, DECaLS (and similar variants).
        maxMagError (float, optional): Objects with magnitude uncertainties greater than this value will be
            cut from retrieved catalogs, subject to the minimum number of bands used by 
            :func:`checkMagErrors`.
    
    Returns:
        Retriever function, a dictionary containing additional parameters that the function needs to be
        given, and the name of the PhotoRedshiftEngine passband set to use. 
        
        Returns `None` if there is no match to `database`.
    
    """
    
    retriever=None
    retrieverOptions=None
    passbandSet='SDSS+Ks'
    if database == 'S82':
        retriever=S82Retriever
        retrieverOptions={'maxMagError': maxMagError}
    elif database == 'SDSSDR7':
        retriever=SDSSDR7Retriever
    elif database == 'SDSSDR8':
        retriever=SDSSDR8Retriever
    elif database == 'SDSSDR10':
        retriever=SDSSDR10Retriever
    elif database == 'SDSSDR12':
        retriever=SDSSDR12Retriever
        retrieverOptions={'maxMagError': maxMagError}
    elif database == 'SDSSDR16':
        retriever=SDSSDR16Retriever
        retrieverOptions={'maxMagError': maxMagError}
    elif database == 'PS1':
        passbandSet='PS1'
        import mastcasjobs
        if not os.environ.get('CASJOBS_WSID'):
            raise Exception("Set CASJOBS_WSID environment variable to use PS1 (see: http://ps1images.stsci.edu/ps1_dr2_query.html - to get your WSID, go to http://mastweb.stsci.edu/ps1casjobs/ChangeDetails.aspx and check your 'profile' page)")
        if not os.environ.get('CASJOBS_PW'):
            raise Exception("Set CASJOBS_PW environment variable to use PS1 (see: http://ps1images.stsci.edu/ps1_dr2_query.html)")
        retriever=PS1Retriever
        retrieverOptions={'maxMagError': maxMagError, 'jobs': mastcasjobs.MastCasJobs(context="PanSTARRS_DR2")}
    elif database == 'DESY3':
        passbandSet='DES'
        import easyaccess as ea
        connection=ea.connect(section = 'dessci')
        retriever=DESY3Retriever
        retrieverOptions={'maxMagError': maxMagError, 'connection': connection}
    elif database == 'DESY3+WISE':
        passbandSet='DES+WISE'
        import easyaccess as ea
        connection=ea.connect(section = 'dessci')
        retriever=DESY3WISERetriever
        retrieverOptions={'maxMagError': maxMagError, 'connection': connection}
    elif database == 'DESDR1':
        import easyaccess as ea
        connection=ea.connect(section = 'desdr')
        retriever=DESDR1Retriever
        retrieverOptions={'maxMagError': maxMagError, 'connection': connection}
    elif database == 'DESDR2':
        import easyaccess as ea
        connection=ea.connect(section = 'desdr')
        retriever=DESDR2Retriever
        retrieverOptions={'maxMagError': maxMagError, 'connection': connection}
    elif database == 'KiDSDR4':
        passbandSet='KiDS-VIKING'
        retriever=KiDSDR4Retriever
        retrieverOptions={'maxMagError': maxMagError}
    elif database == 'ATLASDR4':
        passbandSet='KiDS-VIKING'
        retriever=ATLASDR4Retriever
        retrieverOptions={'maxMagError': maxMagError}
    elif database == 'CFHTLenS':
        retriever=CFHTLenSRetriever
        retrieverOptions={'maxMagError': maxMagError}
    elif database == 'DELVEDR2':
        retriever=DELVEDR2Retriever
        retrieverOptions={'maxMagError': maxMagError}
    elif database == 'DL_DECaLSDR10':
        retriever=DL_DECaLSDR10Retriever
        passbandSet='DECaLS'
        retrieverOptions={'maxMagError': maxMagError}
    elif database.find("DECaLS") != -1 and database != 'DECaLSDR10DL':
        if database == "DECaLS":
            raise Exception("Specify either 'DECaLSDR8', 'DECaLSDR9', or 'DECaLSDR10' instead of DECaLS")
        if database == 'DECaLSDR8':
            DR="DR8"
            retriever=DECaLSDR8Retriever
        elif database == 'DECaLSDR9':
            DR="DR9"
            retriever=DECaLSDR9Retriever
        elif database == "DECaLSDR10":
            DR="DR10"
            retriever=DECaLSDR10Retriever
        else:
            raise Exception("Didn't understand DECaLS database choice")
        passbandSet='DECaLS'
        # For DECaLS, need the bricks files that define survey on the sky
        # These were previously included in zCluster, but now we fetch over web and cache
        bricksCacheDir=CACHE_DIR
        makeCacheDir()
        bricksPath=bricksCacheDir+os.path.sep+"survey-bricks.fits.gz"
        if os.path.exists(bricksPath) == False:
            print("... fetching and caching DECaLS survey-bricks.fits.gz ...")
            urllib.request.urlretrieve("https://portal.nersc.gov/cfs/cosmo/data/legacysurvey/%s/survey-bricks.fits.gz" % (DR.lower()), bricksPath)
        # bricksDRPath=bricksCacheDir+os.path.sep+"survey-bricks-%s-south.fits.gz" % (DR.lower())
        # if os.path.exists(bricksDRPath) == False:
            # print("... fetching and caching DECaLS survey-bricks-%s-south.fits.gz ..." % (DR.lower()))
            # urllib.request.urlretrieve("https://portal.nersc.gov/cfs/cosmo/data/legacysurvey/%s/south/survey-bricks-%s-south.fits.gz" % (DR.lower(), DR.lower()), bricksDRPath)
        bricksTab=atpy.Table().read(bricksPath)
        # DRTab=atpy.Table().read(bricksDRPath)
        # DRTab.rename_column("brickname", "BRICKNAME")
        retrieverOptions={'maxMagError': maxMagError, 'bricksTab': bricksTab}#, 'DRTab': DRTab}
    elif database == 'RubinDP0':
        RSP_TAP_SERVICE='https://data.lsst.cloud/api/tap'
        homeDir=os.path.expanduser('~')
        tokenFileName=os.path.join(homeDir,'.rsp-tap.token')
        if os.path.exists(tokenFileName) == False:
            raise Exception("You need to create file %s containing a valid token string - see https://rsp.lsst.io/guides/auth/creating-user-tokens.html for how to create a token" % (tokenFileName))
        with open(tokenFileName, 'r') as f:
            tokenStr=f.readline()
        cred=pyvo.auth.CredentialStore()
        cred.set_password("x-oauth-basic", tokenStr)
        credential=cred.get("ivo://ivoa.net/sso#BasicAA")
        rspTAP=pyvo.dal.TAPService(RSP_TAP_SERVICE, credential)
        retrieverOptions={'TAP': rspTAP}
        retriever=RubinDP0Retriever
    elif database == 'CFHTDeep':
        retriever=CFHTDeepRetriever
    elif database == 'CFHTWide':
        retriever=CFHTWideRetriever
   
    return retriever, retrieverOptions, passbandSet

#-------------------------------------------------------------------------------------------------------------
def addWISEPhotometry(RADeg, decDeg, catalog, halfBoxSizeDeg = 36.0/60.):
    """This is an option that can be enabled in other retriever functions by adding 'addWISE': True in
    optionsDict. For this to work, the passbandSet used by PhotoRedshiftEngine must also have the WISE bands
    defined. We can probably tidy this up a bit...
       
    """
    
    cacheDir="WISECache"
    if os.path.exists(cacheDir) == False:
        os.makedirs(cacheDir, exist_ok = True)
    
    # Note that using size= in query below is broken (even though it is on IPAC docs)
    outFileName=cacheDir+os.path.sep+"unWISE_%.6f_%.6f.vot" % (RADeg, decDeg)
    gotFileSuccessfully=False
    while gotFileSuccessfully == False:
        if os.path.exists(outFileName) == False:        
            rah, ram, ras=astCoords.decimal2hms(RADeg, ":").split(":")
            dd, dm, ds=astCoords.decimal2dms(decDeg, ":").split(":")
            radiusSize=halfBoxSizeDeg*3600.0#np.degrees(4./astCalc.da(1.0))*3600.0
            # NOTE: URL modified March 2017 as IPAC seems to have changed name from wise_allwise_p3as_psd to allwise_p3as_psd
            url="http://irsa.ipac.caltech.edu/cgi-bin/Gator/nph-query?spatial=cone&catalog=allwise_p3as_psd&objstr=%sh+%sm+%ss+%sd+%sm+%ss&radius=%d&outfmt=3" % (rah, ram, ras, dd, dm, ds, int(round(radiusSize))) 
            urllib.request.urlretrieve(url, outFileName)
        
        try:
            tab=atpy.Table().read(outFileName)
            gotFileSuccessfully=True
        except:
            print("... error reading WISE catalog: retrying download in 1 min ...")
            ##return None
            #IPython.embed()
            #sys.exit()
            os.remove(outFileName)
            time.sleep(60)
    
    # Now need to merge tab with catalog
    matchRadiusDeg=2.0/3600.0
    bands=[]
    for key in catalog[0].keys():
        if key not in ['RADeg', 'decDeg', 'id'] and key.find("Err") == -1:
            bands.append(key)
    catRAs=[]
    catDecs=[]
    for objDict in catalog:
        catRAs.append(objDict['RADeg'])
        catDecs.append(objDict['decDeg'])
        # Defaults: assume not detected
        objDict['w1']=99.0
        objDict['w1Err']=99.0
        objDict['w2']=99.0
        objDict['w2Err']=99.0
    catRAs=np.array(catRAs)
    catDecs=np.array(catDecs)
    objsToAdd=[]
    count=len(catalog)
    for row in tab:
        rDeg=astCoords.calcAngSepDeg(row['ra'], row['dec'], catRAs, catDecs)
        if rDeg.min() < matchRadiusDeg:
            objDict=catalog[np.argmin(rDeg)]
            objDict['w1']=row['w1mpro']+2.699   # Vega -> AB http://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec4_4h.html#conv2ab
            objDict['w1Err']=row['w1sigmpro']
            objDict['w2']=row['w2mpro']+3.339
            objDict['w2Err']=row['w2sigmpro']
        else:
            count=count+1
            objDict={}
            objDict['id']=count
            objDict['RADeg']=row['ra']
            objDict['decDeg']=row['dec']
            objDict['w1']=row['w1mpro']+2.699
            objDict['w1Err']=row['w1sigmpro']
            objDict['w2']=row['w2mpro']+3.339
            objDict['w2Err']=row['w2sigmpro']
            for b in bands:
                objDict[b]=99.0
                objDict['%sErr' % (b)]=99.0
            objsToAdd.append(objDict)
    catalog=catalog+objsToAdd

    return catalog

#-------------------------------------------------------------------------------------------------------------
def S82Retriever(RADeg, decDeg, halfBoxSizeDeg = 20.2/60.0, optionsDict = {}):
    """Retrieves SDSS Stripe 82 photometry at the given position.
    
    """

    makeCacheDir()
    
    # We used PhotoObj for slit masks, because Galaxy misses some stuff, particularly at high-z
    #tableName="PhotoObj"
    tableName="Galaxy"
    
    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
    
    if os.path.exists(cacheDir) == False:
        os.makedirs(cacheDir, exist_ok = True)
        
    url = 'http://cas.sdss.org/stripe82/en/tools/search/x_sql.asp'
            
    outFileName=cacheDir+os.path.sep+"S82_%.4f_%.4f_%.4f_%s.csv" % (RADeg, decDeg, halfBoxSizeDeg, tableName)
    print("... getting SDSS Stripe82 photometry (file: %s) ..." % (outFileName))
            
    if os.path.exists(outFileName) == False:
    
        print("... fetching from the internet ...")

        # Clean galaxy photometry query - note flags for r-band only, may want to change
        # Assuming for limits here on equator, so don't bother with cos(dec)
        #
        # We may want to add something that does multiple queries if we want a bigger area from the standard
        # SDSS query interface
        
        # Less conservative query - seems it's actually the galaxy view that's missing S82 objects
        RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, halfBoxSizeDeg)
        sql="""SELECT ra,dec,dered_u,dered_g,dered_r,dered_i,dered_z,Err_u,Err_g,Err_r,Err_i,Err_z,flags_r,run 
            FROM %s
            WHERE 
            (run=206 OR run=106) AND
            ra BETWEEN %.6f and %.6f AND dec BETWEEN %.6f and %.6f 
            """ % (tableName, RAMin, RAMax, decMin, decMax)
        #print sql
        #AND Err_g < 0.3 AND Err_r < 0.3 AND Err_i < 0.3 
        
        # Old, conservative query that ends up missing a lot of galaxies that we can see in S82 i-band
        #sql="""SELECT ra,dec,dered_u,dered_g,dered_r,dered_i,dered_z,Err_u,Err_g,Err_r,Err_i,Err_z,flags_r,run 
            #FROM Galaxy 
            #WHERE 
            #(run=206 OR run=106) AND
            #ra BETWEEN %.6f and %.6f AND dec BETWEEN %.6f and %.6f 
            #AND ((flags_r & 0x10000000) != 0) 
            #-- detected in BINNED1 
            #AND ((flags_r & 0x8100000c00a0) = 0) 
            #-- not NOPROFILE, PEAKCENTER, NOTCHECKED, PSF_FLUX_INTERP, SATURATED, 
            #-- or BAD_COUNTS_ERROR. 
            #-- if you want to accept objects with interpolation problems for PSF mags, 
            #-- change this to: AND ((flags_r & 0x800a0) = 0) 
            #AND (((flags_r & 0x400000000000) = 0) or (psfmagerr_r <= 0.2)) 
            #-- not DEBLEND_NOPEAK or small PSF error 
            #-- (substitute psfmagerr in other band as appropriate) 
            #AND (((flags_r & 0x100000000000) = 0) or (flags_r & 0x1000) = 0) 
            #-- not INTERP_CENTER or not COSMIC_RAY - omit this AND clause if you want to 
            #-- accept objects with interpolation problems for PSF mags.
            #AND Err_g < 0.2 AND Err_r < 0.2 AND Err_i < 0.2 
            #""" % (RAMin, RAMax, decMin, decMax)
        
        # Bail if no chance this is anywhere near S82
        if decMin > 2.5 or decMax < -2.5:
            print("... not close to S82 region ...")
            return None
        
        # Filter SQL so that it'll work
        fsql = ''
        for line in sql.split('\n'):
            fsql += line.split('--')[0] + ' ' + os.linesep;
        
        params=urllib.parse.urlencode({'cmd': fsql, 'format': "csv"})
        response=None
        while response == None:
            try:
                response=urllib.request.urlopen(url+'?%s' % (params))
            except:
                print("Network down? Waiting 30 sec...")
                time.sleep(30)
            
        lines=response.read()
        lines=lines.decode()
        lines=lines.split("\n")

        outFile=open(outFileName, "w")
        for line in lines:
            outFile.write(line+"\n")
        outFile.close()
    
    else:
        
        inFile=open(outFileName, "r")
        lines=inFile.readlines()
        inFile.close()
    
    # Parse .csv into catalog
    if lines[0].find("No objects have been found") != -1 or len(lines) > 1 and lines[1][:5] == "ERROR":
        catalog=None
    elif lines[1] == 'ERROR\n':
        raise Exception("Error returned by SDSS query - probably too many objects?")
    else:
        catalog=[]
        idCount=0
        for line in lines[1:]: # first line always heading
            if len(line) > 3:
                photDict={}
                idCount=idCount+1
                bits=line.replace("\n", "").split(",")
                photDict['id']=idCount    # just so we have something
                try:
                    photDict['RADeg']=float(bits[0])
                except:
                    if lines[1][:46] == '"ERROR: Maximum 60 queries allowed per minute.':
                        print("... exceeded server queries per minute limit - waiting ...")
                        time.sleep(70)
                        os.remove(outFileName)
                        return "retry"
                    else:
                        print("what?")
                        ipshell()
                        sys.exit()
                photDict['decDeg']=float(bits[1])
                photDict['u']=float(bits[2])
                photDict['g']=float(bits[3])
                photDict['r']=float(bits[4])
                photDict['i']=float(bits[5])
                photDict['z']=float(bits[6])
                photDict['uErr']=float(bits[7])
                photDict['gErr']=float(bits[8])
                photDict['rErr']=float(bits[9])
                photDict['iErr']=float(bits[10])
                photDict['zErr']=float(bits[11])
                photDict['run']=int(bits[13])
                # Apply mag error cuts if given
                # We're just making the mag unconstrained here (missing data), rather than applying a limit
                # If we don't have a minimum of three useful bands, reject
                if 'maxMagError' in list(optionsDict.keys()):
                    keep=checkMagErrors(photDict, optionsDict['maxMagError'])
                else:
                    keep=True
                if keep == True:
                    catalog.append(photDict)
        
        # This guards against us letting through an empty catalog -  e.g. if not in Stripe 82, we get a 
        # different run number, so object doesn't get added to catalog.
        if catalog == []:
            catalog=None
            
    return catalog

#-------------------------------------------------------------------------------------------------------------
def checkMagErrors(photDict, maxMagError, minBands = 3, bands = ['u', 'g', 'r', 'i', 'z']):
    """Checks if the magnitudes in photDict are less than maxMagError, for a minimum number of minBands 
    photometric bands.
    
    Returns True if the photDict passes the test, False if not
    
    """
    
    keep=True
    rejectCount=0
    for band in bands:
        if photDict['%sErr' % (band)] > maxMagError or photDict['%s' % (band)] < 0:
            photDict['%s' % (band)]=99.
            photDict['%sErr' % (band)]=99.
            rejectCount=rejectCount+1
    if len(bands) - rejectCount < minBands:
        keep=False
    
    return keep

#-------------------------------------------------------------------------------------------------------------
def DESY3Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Retrieves DES Y3 photometry at the given position. This assumes you have easyaccess installed
    (https://pypi.python.org/pypi/easyaccess/1.0.7) and access rights for the non-public tables
    in the DES Oracle database.
    
    """
    
    return DESRetriever(RADeg, decDeg, DR = 'Y3', halfBoxSizeDeg = halfBoxSizeDeg, optionsDict = optionsDict)

#-------------------------------------------------------------------------------------------------------------
def DESY3WISERetriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Retrieves DES Y3 photometry joined to AllWISE at the given position. This assumes you have easyaccess 
    installed (https://pypi.python.org/pypi/easyaccess/1.0.7) and access rights for the non-public tables
    in the DES Oracle database.
    
    """
    
    return DESRetriever(RADeg, decDeg, DR = 'Y3+WISE', halfBoxSizeDeg = halfBoxSizeDeg, optionsDict = optionsDict)

#-------------------------------------------------------------------------------------------------------------
def DESDR1Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Retrieves DES DR1 photometry at the given position. This assumes you have easyaccess installed
    (https://pypi.python.org/pypi/easyaccess/1.0.7) and have registered for a login to access the DES
    Oracle database
    
    """
    
    return DESRetriever(RADeg, decDeg, DR = 'DR1', halfBoxSizeDeg = halfBoxSizeDeg, optionsDict = optionsDict)

#-------------------------------------------------------------------------------------------------------------
def DESDR2Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Retrieves DES DR2 photometry at the given position. This assumes you have easyaccess installed
    (https://pypi.python.org/pypi/easyaccess/1.0.7) and have registered for a login to access the DES
    Oracle database
    
    """
    
    return DESRetriever(RADeg, decDeg, DR = 'DR2', halfBoxSizeDeg = halfBoxSizeDeg, optionsDict = optionsDict)

#-------------------------------------------------------------------------------------------------------------
def DESRetriever(RADeg, decDeg, DR = 'DR1', halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Retrieves DES photometry. Assumes you have easyaccess installed
    (https://pypi.python.org/pypi/easyaccess/1.0.7) and the necessary login to access either public or
    proprietary tables in the DES Oracle database.
    
    Use DR = 'DR1' or 'DR2' to access the public photometry, DR = 'Y3' to access the current 'Gold' photometry
    
    """

    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
    
    if os.path.exists(cacheDir) == False:
        os.makedirs(cacheDir, exist_ok = True)
    
    try:
        connection=optionsDict['connection']
    except:
        raise Exception("No DES database connection")
    
    if decDeg > 5:
        print("... outside DES area - skipping ...")
        return None
    
    outFileName=cacheDir+os.path.sep+"DES%s_%.4f_%.4f_%.4f.csv" % (DR, RADeg, decDeg, halfBoxSizeDeg)      

    # This bit has to be outside, in case we load data from cache
    if DR == 'Y3' or DR == 'Y3+WISE':
        magKey="SOF_CM_MAG_CORRECTED_$BAND"
        magErrKey="SOF_CM_MAG_ERR_$BAND"
    elif DR == 'DR1' or DR == 'DR2':
        magKey="MAG_AUTO_$BAND_DERED"
        magErrKey="MAGERR_AUTO_$BAND"
    else:
        raise Exception("didn't recognise requested DES DR")

    if os.path.exists(outFileName) == False or 'refetch' in list(optionsDict.keys()) and optionsDict['refetch'] == True:
        RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, halfBoxSizeDeg)
        # Y3 Gold v1.0
        #query="SELECT coadd_object_id, ra, dec, ngmix_cm_mag_g - (3.186 * EBV_SFD98) AS cm_mag_g, ngmix_cm_mag_r - (2.140 * EBV_SFD98) AS cm_mag_r, ngmix_cm_mag_i - (1.569 * EBV_SFD98) AS cm_mag_i, ngmix_cm_mag_z - (1.196 * EBV_SFD98) AS cm_mag_z, ngmix_cm_mag_err_g AS cm_mag_err_g, ngmix_cm_mag_err_r AS cm_mag_err_r, ngmix_cm_mag_err_i AS cm_mag_err_i, ngmix_cm_mag_err_z AS cm_mag_err_z FROM y3_gold_1_0 SAMPLE WHERE ra BETWEEN %.6f and %.6f AND dec BETWEEN %.6f and %.6f AND flag_gold = 0 AND flag_footprint = 1 AND flag_foreground = 0 AND extended_class_mash BETWEEN 3 AND 4" % (RAMin, RAMax, decMin, decMax) 
        # Y3 Gold v2.2
        if DR == 'Y3':
            query="SELECT COADD_OBJECT_ID, RA, DEC, DNF_ZMC_SOF, BPZ_ZMC_SOF, SOF_CM_MAG_CORRECTED_G, SOF_CM_MAG_CORRECTED_R, SOF_CM_MAG_CORRECTED_I, SOF_CM_MAG_CORRECTED_Z, SOF_CM_MAG_ERR_G, SOF_CM_MAG_ERR_R, SOF_CM_MAG_ERR_I, SOF_CM_MAG_ERR_Z FROM Y3_GOLD_2_2 WHERE FLAGS_FOOTPRINT = 1 and FLAGS_FOREGROUND = 0 and bitand(FLAGS_GOLD, 62) = 0 and EXTENDED_CLASS_MASH_SOF = 3 and SOF_CM_MAG_I between 16 and 24 AND RA BETWEEN %.6f AND %.6f AND DEC BETWEEN %.6f and %.6f" % (RAMin, RAMax, decMin, decMax)
        elif DR == 'DR1' or DR == 'DR2':
            query="SELECT RA, DEC, MAG_AUTO_G_DERED, MAG_AUTO_R_DERED, MAG_AUTO_I_DERED, MAG_AUTO_Z_DERED, MAGERR_AUTO_G, MAGERR_AUTO_R, MAGERR_AUTO_I, MAGERR_AUTO_Z FROM %s_MAIN WHERE WAVG_SPREAD_MODEL_I + 3.0*WAVG_SPREADERR_MODEL_I > 0.005 and WAVG_SPREAD_MODEL_I + 1.0*WAVG_SPREADERR_MODEL_I > 0.003 and WAVG_SPREAD_MODEL_I - 1.0*WAVG_SPREADERR_MODEL_I > 0.001 and WAVG_SPREAD_MODEL_I > -1 and IMAFLAGS_ISO_I = 0 and MAG_AUTO_I < 24 and RA BETWEEN %.6f AND %.6f AND DEC BETWEEN %.6f and %.6f" % (DR, RAMin, RAMax, decMin, decMax)
        elif DR == 'Y3+WISE':
            query="SELECT D.COADD_OBJECT_ID, D.RA, D.DEC, DNF_ZMC_SOF, BPZ_ZMC_SOF, SOF_CM_MAG_CORRECTED_G, SOF_CM_MAG_CORRECTED_R, SOF_CM_MAG_CORRECTED_I, SOF_CM_MAG_CORRECTED_Z, SOF_CM_MAG_ERR_G, SOF_CM_MAG_ERR_R, SOF_CM_MAG_ERR_I, SOF_CM_MAG_ERR_Z, W1MPRO, W1SIGMPRO, W2MPRO, W2SIGMPRO FROM Y3_GOLD_2_2 D LEFT OUTER JOIN DES_ADMIN.Y3A2_WISE_DES W ON D.COADD_OBJECT_ID = W.COADD_OBJECT_ID WHERE FLAGS_FOOTPRINT = 1 and FLAGS_FOREGROUND = 0 and bitand(FLAGS_GOLD, 62) = 0 and EXTENDED_CLASS_MASH_SOF = 3 and SOF_CM_MAG_I between 16 and 24 AND D.RA BETWEEN %.6f AND %.6f AND D.DEC BETWEEN %.6f and %.6f" % (RAMin, RAMax, decMin, decMax)
        else:
            raise Exception("didn't recognise requested DES DR")

        # This should restart the connection if it drops
        if connection.ping() == False:
            print("... DES database connection lost: reconnecting ...")
            import easyaccess as ea
            connection=ea.connect()
        if connection.ping() == True:
            connection.query_and_save(query, outFileName)
        else:
            raise Exception("DES database connection lost")

    # No output is written if the query returns nothing... so we'll write a blank file to save repeating the query next time
    catalog=[]
    if os.path.exists(outFileName) == False:
        outFile=open(outFileName, "w")
        outFile.write("# No objects returned by query\n")
        outFile.close()
    else:
        
        tab=atpy.Table().read(outFileName, format = 'csv')

        idCount=0
        for row in tab:
            idCount=idCount+1
            photDict={}
            photDict['id']=idCount    # just so we have something - we could use COADD_OBJECT_ID but skipping for now
            photDict['RADeg']=row['RA']
            photDict['decDeg']=row['DEC']
            photDict['g']=row[magKey.replace("$BAND", "G")]
            photDict['r']=row[magKey.replace("$BAND", "R")]
            photDict['i']=row[magKey.replace("$BAND", "I")]
            photDict['z']=row[magKey.replace("$BAND", "Z")]
            photDict['gErr']=row[magErrKey.replace("$BAND", "G")]
            photDict['rErr']=row[magErrKey.replace("$BAND", "R")]
            photDict['iErr']=row[magErrKey.replace("$BAND", "I")]
            photDict['zErr']=row[magErrKey.replace("$BAND", "Z")]
            if DR.find("+WISE") != -1:
                # Convert to AB mags and handle missing values
                if row['W1MPRO'] < 0:
                    w1Mag=99.0
                    w1MagErr=99.0
                else:
                    w1Mag=row['W1MPRO']+2.699
                    w1MagErr=row['W1SIGMPRO']
                if row['W2MPRO'] < 0:
                    w2Mag=99.0
                    w2MagErr=99.0
                else:
                    w2Mag=row['W2MPRO']+3.339
                    w2MagErr=row['W2SIGMPRO']
                photDict['w1']=w1Mag
                photDict['w1Err']=w1MagErr
                photDict['w2']=w2Mag
                photDict['w2Err']=w2MagErr
            
            # Apply mag error cuts if given
            # For PS1, missing values are -999 - our current checkMagErrors routine will fish those out
            # We're just making the mag unconstrained here (missing data), rather than applying a limit
            # If we don't have a minimum of three useful bands, reject
            if 'maxMagError' in list(optionsDict.keys()):
                keep=checkMagErrors(photDict, optionsDict['maxMagError'], bands = ['g', 'r', 'i', 'z'])
            else:
                keep=True
            
            ## Additional colour cuts (as used in some DES papers - see e.g., splashback draft)
            #gr=photDict['g']-photDict['r']
            #ri=photDict['r']-photDict['i']
            #iz=photDict['i']-photDict['z']
            #if keep == True and -1 < gr < 3 and -1 < ri < 2.5 and -1 < iz < 2:
                #keep=True
            #else:
                #keep=False
                
            if keep == True:
                catalog.append(photDict)
    
    #if 'addWISE' in optionsDict.keys() and optionsDict['addWISE'] == True:
        #catalog=addWISEPhotometry(RADeg, decDeg, catalog, halfBoxSizeDeg = halfBoxSizeDeg)
        
    return catalog

#-------------------------------------------------------------------------------------------------------------
def KiDSDR4Retriever(RADeg, decDeg, halfBoxSizeDeg = 18.0/60.0, optionsDict = {}):
    """Retrieves KiDS DR4 photometry (which includes VIKING IR bands).
    
    """
    
    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
    
    if os.path.exists(cacheDir) == False:
        os.makedirs(cacheDir, exist_ok = True)
        
    # This defines the rough final KIDS area (split into two fields)
    inKIDSRegion=False
    if RADeg > 120 and RADeg < 240 and decDeg > -5 and decDeg < 5:
        inKIDSRegion=True
    if RADeg > 330 and decDeg > -37 and decDeg < -25:
        inKIDSRegion=True
    if RADeg < 60 and decDeg > -37 and decDeg < -25:
        inKIDSRegion=True
    if inKIDSRegion == False:
        print("... outside KIDS area - skipping ...")
        return None

    outFileName=cacheDir+os.path.sep+"KiDSDR4_%.4f_%.4f_%.4f.fits" % (RADeg, decDeg, halfBoxSizeDeg)      
    if os.path.exists(outFileName) == False or 'refetch' in list(optionsDict.keys()) and optionsDict['refetch'] == True:
        RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, halfBoxSizeDeg)
        print("... downloading catalog %s ..." % (outFileName))
        query="select id, RAJ2000, DECJ2000, MAG_GAAP_u, MAG_GAAP_g, MAG_GAAP_r, MAG_GAAP_i, MAG_GAAP_Z, MAG_GAAP_Y, MAG_GAAP_J, MAG_GAAP_H, MAG_GAAP_Ks, MAGERR_GAAP_u, MAGERR_GAAP_g, MAGERR_GAAP_r, MAGERR_GAAP_i, MAGERR_GAAP_Z, MAGERR_GAAP_Y, MAGERR_GAAP_J, MAGERR_GAAP_H, MAGERR_GAAP_Ks, SG2DPHOT from KiDS_DR4_0_ugriZYJHKs_cat_fits_V3 where RAJ2000 BETWEEN %.6f AND %.6f AND DECJ2000 BETWEEN %.6f and %.6f and SG2DPHOT = 0;" % (RAMin, RAMax, decMin, decMax)        
        r=requests.get('http://archive.eso.org/tap_cat/sync?', 
                        params = {'REQUEST': 'doQuery',
                                  'LANG': 'ADQL',
                                  'MAXREC': 1000000,
                                  'FORMAT': 'fits',
                                  'QUERY': query},
                        stream = True)
        with open(outFileName, 'wb') as outFile: 
            r.raw.decode_content = True
            outFile.write(r.raw.data) 
    
    # Load/parse table
    print("... reading catalog %s ..." % (outFileName))
    try:
        tab=atpy.Table().read(outFileName)
    except ValueError:
        print("... no objects in catalog - skipping ...")
        return None
    
    magKey="MAG_GAAP_$BAND"
    magErrKey="MAGERR_GAAP_$BAND"
    
    # First, get rid of nans or nonsensical values
    bands=['u', 'g', 'r', 'i', 'Z', 'Y', 'J', 'H', 'Ks']
    for b in bands:
        #tab=tab[np.where(np.isnan(tab[magKey.replace("$BAND", b)]) == False)]
        #tab=tab[np.where(np.isnan(tab[magErrKey.replace("$BAND", b)]) == False)]
        tab=tab[np.where(tab[magKey.replace("$BAND", b)] > 0)]
        tab=tab[np.where(tab[magErrKey.replace("$BAND", b)] > 0)]
        
    # NOTE: Need to add JHKs mags
    idCount=0
    catalog=[]
    for row in tab:
        idCount=idCount+1
        photDict={}
        photDict['id']=idCount    # just so we have something - we could use ID but skipping for now
        photDict['RADeg']=row['RAJ2000']
        photDict['decDeg']=row['DECJ2000']
        # KiDS DR4 mags we're using are already extinction corrected
        photDict['u']=row[magKey.replace("$BAND", "u")]#-row['EXT_SFD_U']
        photDict['g']=row[magKey.replace("$BAND", "g")]#-row['EXT_SFD_G']
        photDict['r']=row[magKey.replace("$BAND", "r")]#-row['EXT_SFD_R']
        photDict['i']=row[magKey.replace("$BAND", "i")]#-row['EXT_SFD_I']
        photDict['Z']=row[magKey.replace("$BAND", "Z")]#-row['EXT_SFD_U']
        photDict['Y']=row[magKey.replace("$BAND", "Y")]#-row['EXT_SFD_G']
        photDict['J']=row[magKey.replace("$BAND", "J")]#-row['EXT_SFD_R']
        photDict['H']=row[magKey.replace("$BAND", "H")]#-row['EXT_SFD_I']
        photDict['Ks']=row[magKey.replace("$BAND", "Ks")]#-row['EXT_SFD_I']
        photDict['uErr']=row[magErrKey.replace("$BAND", "u")]
        photDict['gErr']=row[magErrKey.replace("$BAND", "g")]
        photDict['rErr']=row[magErrKey.replace("$BAND", "r")]
        photDict['iErr']=row[magErrKey.replace("$BAND", "i")]
        photDict['ZErr']=row[magErrKey.replace("$BAND", "Z")]
        photDict['YErr']=row[magErrKey.replace("$BAND", "Y")]
        photDict['JErr']=row[magErrKey.replace("$BAND", "J")]
        photDict['HErr']=row[magErrKey.replace("$BAND", "H")]
        photDict['KsErr']=row[magErrKey.replace("$BAND", "Ks")]
        if 'maxMagError' in list(optionsDict.keys()):
            keep=checkMagErrors(photDict, optionsDict['maxMagError'], bands = ['u', 'g', 'r', 'i'])
        else:
            keep=True
                    
        if keep == True:
            catalog.append(photDict)
        
    return catalog
    
#-------------------------------------------------------------------------------------------------------------
def ATLASDR4Retriever(RADeg, decDeg, halfBoxSizeDeg = 18.0/60.0, optionsDict = {}):
    """Retrieves VST ATLAS DR4 photometry via ESO.
    
    """

    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
    
    if os.path.exists(cacheDir) == False:
        os.makedirs(cacheDir, exist_ok = True)
        
    outFileName=cacheDir+os.path.sep+"ATLASDR4_%.4f_%.4f_%.4f.fits" % (RADeg, decDeg, halfBoxSizeDeg)          
    if os.path.exists(outFileName) == False or 'refetch' in list(optionsDict.keys()) and optionsDict['refetch'] == True:
        RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, halfBoxSizeDeg)
        print("... downloading catalog %s ..." % (outFileName)) 
        query="select sourceID, ra2000, dec2000, uPetroMag, uPetroMagErr, gPetroMag, gPetroMagErr, rPetroMag, rPetroMagErr, iPetroMag, iPetroMagErr, zPetroMag, zPetroMagErr, uAperMagNoAperCorr3, uAperMag3Err, gAperMagNoAperCorr3, gAperMag3Err, rAperMagNoAperCorr3, rAperMag3Err, iAperMagNoAperCorr3, iAperMag3Err, zAperMagNoAperCorr3, zAperMag3Err, aU, aG, aR, aI, aZ, pGalaxy, pStar, pNoise, pSaturated from atlas_er4_ugriz_catMetaData_fits_V3 where ra2000 BETWEEN %.4f AND %.4f AND dec2000 BETWEEN %.4f and %.4f and priOrSec = 0;" % (RAMin, RAMax, decMin, decMax)   
        r=requests.get('http://archive.eso.org/tap_cat/sync?', 
                        params = {'REQUEST': 'doQuery',
                                  'LANG': 'ADQL',
                                  'MAXREC': 1000000,
                                  'FORMAT': 'fits',
                                  'QUERY': query},
                        stream = True)
        with open(outFileName, 'wb') as outFile: 
            r.raw.decode_content = True
            outFile.write(r.raw.data) 
            
    # Load/parse table
    print("... reading catalog %s ..." % (outFileName))
    tab=atpy.Table().read(outFileName)
    if len(tab) == 0:
        print("... no objects in catalog - skipping ...")
        return None

    magKey="$BANDPETROMAG"
    magErrKey="$BANDPETROMAGERR"
    #magKey="$BANDAPERMAGNOAPERCORR3"
    #magErrKey="$BANDAPERMAG3ERR"
    
    # First, get rid of nans or nonsensical values
    # We apply extinction correction in-place here
    # NOTE: Zapping u (most uncertain with regards photo calib I think) makes no significant difference
    bands=['u', 'g', 'r', 'i', 'z']
    for b in bands:
        tab[magKey.replace("$BAND", b.upper())][np.isnan(tab[magKey.replace("$BAND", b.upper())])]=99.0
        tab[magErrKey.replace("$BAND", b.upper())][np.isnan(tab[magErrKey.replace("$BAND", b.upper())])]=99.0
        tab[magKey.replace("$BAND", b.upper())]=tab[magKey.replace("$BAND", b.upper())]-tab['A%s' % (b.upper())]
        
    # Transform to SDSS in-place - maybe this works better?
    # These transforms are from doing some algebra on the ones in the VST-ATLAS paper and throwing away ~0.00x terms
    # NOTE: These don't work at all well (much worse than not doing this)
    #iSDSS=tab[magKey.replace("$BAND", 'I')]-0.025
    #zSDSS=(tab[magKey.replace("$BAND", 'Z')]-0.04*tab[magKey.replace("$BAND", 'I')]+0.04)/0.96
    #rSDSS=(tab[magKey.replace("$BAND", 'R')]+0.0316*tab[magKey.replace("$BAND", 'G')])/1.03
    #gSDSS=(tab[magKey.replace("$BAND", 'G')]/0.95)-0.048*(tab[magKey.replace("$BAND", 'R')]+0.316*tab[magKey.replace("$BAND", 'G')])-0.063
    #uSDSS=1.01*(tab[magKey.replace("$BAND", 'U')]-0.01*gSDSS+0.27)
    #tab[magKey.replace("$BAND", 'U')]=uSDSS
    #tab[magKey.replace("$BAND", 'G')]=gSDSS
    #tab[magKey.replace("$BAND", 'R')]=rSDSS
    #tab[magKey.replace("$BAND", 'I')]=iSDSS
    #tab[magKey.replace("$BAND", 'Z')]=zSDSS
    
    # Empirical corrections (based on own comparison with SDSS galaxy photometry for small subsample)
    # Again, these don't help at all really
    #tab[magKey.replace("$BAND", 'U')]=tab[magKey.replace("$BAND", 'U')]-0.286
    #tab[magKey.replace("$BAND", 'G')]=tab[magKey.replace("$BAND", 'G')]-0.080
    #tab[magKey.replace("$BAND", 'R')]=tab[magKey.replace("$BAND", 'R')]-0.009
    #tab[magKey.replace("$BAND", 'I')]=tab[magKey.replace("$BAND", 'I')]-0.084
    #tab[magKey.replace("$BAND", 'Z')]=tab[magKey.replace("$BAND", 'Z')]-0.073
    
    # Classification cuts
    tab=tab[np.where(tab['PGALAXY'] > 0.5)]
    tab=tab[np.where(tab['PNOISE'] < 0.01)]
    tab=tab[np.where(tab['PSATURATED'] < 0.01)]
    
    idCount=0
    catalog=[]
    for row in tab:
        idCount=idCount+1
        photDict={}
        photDict['id']=idCount    # just so we have something - we could use ID but skipping for now
        photDict['RADeg']=row['RA2000']
        photDict['decDeg']=row['DEC2000']
        for b in bands:
            photDict[b]=row[magKey.replace("$BAND", b.upper())]
            photDict[b+'Err']=row[magErrKey.replace("$BAND", b.upper())]
        if 'maxMagError' in list(optionsDict.keys()):
            keep=checkMagErrors(photDict, optionsDict['maxMagError'], bands = bands)
        else:
            keep=True
        if keep == True:
            catalog.append(photDict)
        
    return catalog

#------------------------------------------------------------------------------------------------------------
def fixcolnames(tab):
    """Fix column names returned by the casjobs query. Needed by PS1Retriever.
    
    Parameters
    ----------
    tab (astropy.table.Table): Input table

    Returns reference to original table with column names modified"""

    pat = re.compile(r'\[(?P<name>[^[]+)\]')
    for c in tab.colnames:
        m = pat.match(c)
        if not m:
            raise ValueError("Unable to parse column name '{}'".format(c))
        newname = m.group('name')
        tab.rename_column(c,newname)
    return tab

#-------------------------------------------------------------------------------------------------------------
def PS1Retriever(RADeg, decDeg, halfBoxSizeDeg = 18.0/60.0, optionsDict = {}):
    """Retrieves PS1 photometry at the given position.
        
    """
    
    import mastcasjobs

    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
    
    if os.path.exists(cacheDir) == False:
        os.makedirs(cacheDir, exist_ok = True)

    outFileName=cacheDir+os.path.sep+"PS1_%.4f_%.4f_%.4f.fits" % (RADeg, decDeg, halfBoxSizeDeg)      
    print("... getting PS1 photometry (file: %s) ..." % (outFileName))

    if decDeg < -30:
        print("... outside PS1 area - skipping ...")
        return None
    
    if os.path.exists(outFileName) == False or 'refetch' in list(optionsDict.keys()) and optionsDict['refetch'] == True:
        print("... fetching from the internet ...")
        #query="""select o.objID, o.raMean, o.decMean,
        #m.gMeanKronMag, m.rMeanKronMag, m.iMeanKronMag, m.zMeanKronMag, m.yMeanKronMag, m.gMeanKronMagErr, m.rMeanKronMagErr, m.iMeanKronMagErr, m.zMeanKronMagErr, m.yMeanKronMagErr, m.iMeanPSFMag
        #from fGetNearbyObjEq(%.6f, %.6f, %.6f) nb
        #inner join ObjectThin o on o.objid=nb.objid and o.nDetections > 3
        #inner join MeanObject m on o.objid=m.objid and o.uniquePspsOBid=m.uniquePspsOBid""" % (RADeg, decDeg, halfBoxSizeDeg*60)
        RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, halfBoxSizeDeg)
        #query="""select m.objID, m.raMean, m.decMean, m.gKronMag, m.rKronMag, m.iKronMag, m.zKronMag, m.yKronMag, m.gKronMagErr, m.rKronMagErr, m.iKronMagErr, m.zKronMagErr, m.yKronMagErr, m.iPSFMag from StackObjectView m where m.raMean > %.6f and m.raMean < %.6f and m.decMean > %.6f and m.decMean < %.6f""" % (RAMin, RAMax, decMin, decMax)
        #query="""select objID, raMean, decMean, gKronMag, rKronMag, iKronMag, zKronMag, yKronMag, gKronMagErr, rKronMagErr, iKronMagErr, zKronMagErr, yKronMagErr, iPSFMag, primaryDetection from StackObjectView where raMean > %.6f and raMean < %.6f and decMean > %.6f and decMean < %.6f""" % (RAMin, RAMax, decMin, decMax)
        query="""select objID, raMean, decMean, gApMag, rApMag, iApMag, zApMag, yApMag, gApMagErr, rApMagErr, iApMagErr, zApMagErr, yApMagErr, iPSFMag, primaryDetection from StackObjectView where raMean > %.6f and raMean < %.6f and decMean > %.6f and decMean < %.6f""" % (RAMin, RAMax, decMin, decMax)
        #query="""select o.objID, o.raMean, o.decMean,
        #m.gKronMag, m.rKronMag, m.iKronMag, m.zKronMag, m.yKronMag, m.gKronMagErr, m.rKronMagErr, m.iKronMagErr, m.zKronMagErr, m.yKronMagErr, m.iPSFMag
        #from fGetNearbyObjEq(%.6f, %.6f, %.6f) nb
        #inner join ObjectThin o on o.objid=nb.objid and o.nDetections > 3
        #inner join StackObject m on o.objid=m.objid and o.uniquePspsOBid=m.uniquePspsOBid""" % (RADeg, decDeg, halfBoxSizeDeg*60)
        jobs=optionsDict['jobs']
        #jobs=mastcasjobs.MastCasJobs(context="PanSTARRS_DR2")
        tab=jobs.quick(query, task_name="python cone search")
        tab.write(outFileName, overwrite = True)
    else:
        tab=atpy.Table().read(outFileName)

    # Parse table into catalog
    if len(tab) == 0:
        catalog=None
    else:
        # Star-galaxy cut (see: https://outerspace.stsci.edu/display/PANSTARRS/How+to+separate+stars+and+galaxies)
        #tab=tab[(tab['iPSFMag']-tab['iKronMag']) > 0.05]
        tab=tab[np.array(tab['primaryDetection'], dtype = int) == 1] # Because bytes, not int
        EBMinusV=getEBMinusV(RADeg, decDeg, optionsDict = optionsDict) # assume same across field
        catalog=[]
        idCount=0
        bands=['g', 'r', 'i', 'z', 'y']
        for row in tab:
            idCount=idCount+1
            photDict={}
            photDict['id']=idCount    # just so we have something - we could use PS1 ID but skipping for now
            photDict['RADeg']=row['raMean']
            photDict['decDeg']=row['decMean']
            for b in bands:
                photDict[b]=row['%sApMag' % (b)]
                photDict['%sErr' % (b)]=row['%sApMagErr' %  (b)]

            # Correct for dust extinction
            # Taken from: http://www.mso.anu.edu.au/~brad/filters.html
            photDict['g']=photDict['g']-EBMinusV*3.322
            photDict['r']=photDict['r']-EBMinusV*2.544 
            photDict['i']=photDict['i']-EBMinusV*2.265 
            photDict['z']=photDict['z']-EBMinusV*1.846 
            photDict['y']=photDict['y']-EBMinusV*1.570 

            # Apply mag error cuts if given
            # For PS1, missing values are -999 - our current checkMagErrors routine will fish those out
            # We're just making the mag unconstrained here (missing data), rather than applying a limit
            # If we don't have a minimum of three useful bands, reject
            if 'maxMagError' in list(optionsDict.keys()):
                keep=checkMagErrors(photDict, optionsDict['maxMagError'], bands = ['g', 'r', 'i', 'z', 'y'], minBands = 3)
            else:
                keep=True
                
            if keep == True:
                catalog.append(photDict)
        
    return catalog

#-------------------------------------------------------------------------------------------------------------
def SDSSRetriever(RADeg, decDeg, halfBoxSizeDeg = 18.0/60.0, DR = 7, optionsDict = {}):
    """Retrieves SDSS main photometry at the given position.
    
    """
    
    # PhotoPrimary avoids star-galaxy classification problems at higher z, which we need when making slit masks
    # BUT contamination from stars etc. increases scatter in (z_spec - z_phot) by a fair bit (outliers)
    #tableName="PhotoPrimary"
    tableName="Galaxy"
    
    makeCacheDir()
    
    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
    
    if os.path.exists(cacheDir) == False:
        os.makedirs(cacheDir, exist_ok = True)
    
    if DR == 7:
        url='http://cas.sdss.org/astrodr7/en/tools/search/x_sql.asp'
        outFileName=cacheDir+os.path.sep+"SDSSDR7_%.4f_%.4f_%.4f.csv" % (RADeg, decDeg, halfBoxSizeDeg)
        lineSkip=1
    elif DR == 8:
        url='http://skyserver.sdss3.org/dr8/en/tools/search/x_sql.asp'
        outFileName=cacheDir+os.path.sep+"SDSSDR8_%.4f_%.4f_%.4f.csv" % (RADeg, decDeg, halfBoxSizeDeg)
        lineSkip=1
    elif DR == 10:      
        url='http://skyserver.sdss3.org/dr10/en/tools/search/x_sql.aspx'
        outFileName=cacheDir+os.path.sep+"SDSSDR10_%.4f_%.4f_%.4f.csv" % (RADeg, decDeg, halfBoxSizeDeg)
        lineSkip=2
    elif DR == 12:
        # For some reason, SDSS changed their whole web API in ~May 2016 without calling it a new DR
        #url='http://skyserver.sdss.org/dr12/en/tools/search/x_sql.aspx'        
        #url='http://skyserver.sdss.org/dr12/en/tools/search/x_results.aspx'
        url='https://skyserver.sdss.org/dr12/en/tools/search/x_results.aspx?searchtool=SQL&TaskName=Skyserver.Search.SQL&syntax=NoSyntax&ReturnHtml=false&'
        outFileName=cacheDir+os.path.sep+"SDSSDR12_%.4f_%.4f_%.4f.csv" % (RADeg, decDeg, halfBoxSizeDeg)
        lineSkip=2
    elif DR == 16:
        # DR12 server has some fatal error (not at our end, according to error message) so trying this
        url='https://skyserver.sdss.org/dr16/en/tools/search/x_results.aspx?searchtool=SQL&TaskName=Skyserver.Search.SQL&syntax=NoSyntax&ReturnHtml=false&'
        outFileName=cacheDir+os.path.sep+"SDSSDR16_%.4f_%.4f_%.4f.csv" % (RADeg, decDeg, halfBoxSizeDeg)
        lineSkip=2
   
    outFileName=outFileName.replace(".csv", "_%s.csv" % (tableName))
                                    
    print("... getting SDSS DR%d photometry (file: %s) ..." % (DR, outFileName))
        
    if os.path.exists(outFileName) == False or 'refetch' in list(optionsDict.keys()) and optionsDict['refetch'] == True:
        
        print("... fetching from the internet ...")
        
        # Clean galaxy photometry query - note flags for r-band only, may want to change
        # We may want to add something that does multiple queries if we want a bigger area from the standard
        # SDSS query interface
        # NOTE: disabled the 'clean' photometry flags on 21/09/2015
        RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, halfBoxSizeDeg)
        # Sanity check: if RAMin, RAMax have order reversed, SDSS gives us nothing
        #if RAMin > RAMax:
            #print "RAMin, RAMax reversed"
            #IPython.embed()
            #sys.exit()
            #newRAMax=RAMin
            #RAMin=RAMax
            #RAMax=newRAMax
        #if decMin > decMax:
            #print "decMin, decMax reversed"
            #IPython.embed()
            #sys.exit()
            #newDecMax=decMin
            #decMin=decMax
            #decMax=newDecMax
        # PhotoPrimary is like PhotoObj but without multiple matches
        sql="""SELECT ra,dec,dered_u,dered_g,dered_r,dered_i,dered_z,Err_u,Err_g,Err_r,Err_i,Err_z,flags_r,run 
            FROM %s
            WHERE 
            ra BETWEEN %.6f and %.6f AND dec BETWEEN %.6f and %.6f 
            -- AND ((flags_r & 0x10000000) != 0) 
            -- detected in BINNED1 
            -- AND ((flags_r & 0x8100000c00a0) = 0) 
            -- not NOPROFILE, PEAKCENTER, NOTCHECKED, PSF_FLUX_INTERP, SATURATED, 
            -- or BAD_COUNTS_ERROR. 
            -- if you want to accept objects with interpolation problems for PSF mags, 
            -- change this to: AND ((flags_r & 0x800a0) = 0) 
            -- AND (((flags_r & 0x400000000000) = 0) or (psfmagerr_r <= 0.2)) 
            -- not DEBLEND_NOPEAK or small PSF error 
            -- (substitute psfmagerr in other band as appropriate) 
            -- AND (((flags_r & 0x100000000000) = 0) or (flags_r & 0x1000) = 0) 
            -- not INTERP_CENTER or not COSMIC_RAY - omit this AND clause if you want to 
            -- accept objects with interpolation problems for PSF mags.
            --AND Err_g < 0.5 AND Err_r < 0.5 AND Err_i < 0.5 
            """ % (tableName, RAMin, RAMax, decMin, decMax)
    
        # Filter SQL so that it'll work
        fsql = ''
        for line in sql.split('\n'):
            fsql += line.split('--')[0] + ' ' + os.linesep;
    
        params=urllib.parse.urlencode({'cmd': fsql, 'format': "csv"})
        response=None
        while response == None:
            try:
                response=urllib.request.urlopen(url+'%s' % (params))
            except:
                print("Network down? Waiting 30 sec... - if this persists, probably the query URL has changed.")
                time.sleep(30)

        # Some faffing about here because of python2 -> python3 
        lines=response.read()
        lines=lines.splitlines()        
        outFile=open(outFileName, "w")
        strLines=[]
        for line in lines:
            strLines.append(line.decode('utf-8'))
            outFile.write(strLines[-1]+"\n")
        outFile.close()
        lines=strLines
    
    else:
        
        inFile=open(outFileName, "r")
        lines=inFile.readlines()
        inFile.close()
    
    # Parse .csv into catalog
    if lines[0].find("No objects have been found") != -1 or len(lines) > 1 and lines[1][:5] == "ERROR":
        catalog=None
    else:
        catalog=[]
        idCount=0
        for line in lines[lineSkip:]: # first line or two (depending on DR) always heading
            if len(line) > 3:
                photDict={}
                idCount=idCount+1
                bits=line.replace("\n", "").split(",")
                photDict['id']=idCount    # just so we have something
                try:
                    photDict['RADeg']=float(bits[0])
                except:
                    print("... problem with file %s - removing and retrying in 70 sec ..." % (outFileName))
                    os.remove(outFileName)
                    time.sleep(70)
                    return "retry"                    
                    #if lines[1][:46] == '"ERROR: Maximum 60 queries allowed per minute.':
                        #print "... exceeded server queries per minute limit - waiting ..."
                        #time.sleep(70)
                        #os.remove(outFileName)
                        #return "retry"
                    #else:
                        #if line.find("<html>") != -1:
                            #time.sleep(70)
                            #os.remove(outFileName)
                            #return "retry"
                        #else:
                            #print "what?"
                            #IPython.embed()
                            #sys.exit()
                photDict['decDeg']=float(bits[1])
                photDict['u']=float(bits[2])
                photDict['g']=float(bits[3])
                photDict['r']=float(bits[4])
                photDict['i']=float(bits[5])
                photDict['z']=float(bits[6])
                photDict['uErr']=float(bits[7])
                photDict['gErr']=float(bits[8])
                photDict['rErr']=float(bits[9])
                photDict['iErr']=float(bits[10])
                photDict['zErr']=float(bits[11])
                photDict['run']=int(bits[13])
                # Apply mag error cuts if given
                # We're just making the mag unconstrained here (missing data), rather than applying a limit
                # If we don't have a minimum of three useful bands, reject
                if 'maxMagError' in list(optionsDict.keys()):
                    keep=checkMagErrors(photDict, optionsDict['maxMagError'])
                else:
                    keep=True
                if keep == True:
                    catalog.append(photDict)
        
    return catalog

#-------------------------------------------------------------------------------------------------------------
def DECaLSRetriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, DR = None, optionsDict = {}):
    """Retrieves DECaLS DRx tractor catalogs (if they exist) at the given position. Cuts the catalog to the
    radius specified by halfBoxSizeDeg.

    """

    makeCacheDir()
    
    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
    
    if os.path.exists(cacheDir) == False:
        os.makedirs(cacheDir, exist_ok = True)

    # Organised such that after this, have subdir with degrees RA (e.g. 000/ is 0 < RADeg < 1 degree)
    #basePath="https://portal.nersc.gov/project/cosmo/data/legacysurvey/dr8/south/tractor/"
    basePath="https://portal.nersc.gov/project/cosmo/data/legacysurvey/%s/south/tractor/" % (DR.lower())

    outFileName=cacheDir+os.path.sep+"DECaLS%s_%.4f_%.4f_%.2f.fits" % (DR, RADeg, decDeg,
                                                                            halfBoxSizeDeg)
    
    print("... getting DECaLS %s photometry ..." % (DR))

    bricksTab=optionsDict['bricksTab']
    # DRTab=optionsDict['DRTab']

    # Find matching tractor catalogs and download/cache .fits table files
    # Previously this would fail for very small search areas - hence added centre coords search
    # NOTE: We've doubled asked for halfBoxSizeDeg here to get all the matching tiles
    # Otherwise we tend to miss...
    RAMask=np.logical_and(np.greater(RADeg, bricksTab['RA1']), np.less(RADeg, bricksTab['RA2']))
    decMask=np.logical_and(np.greater(decDeg, bricksTab['DEC1']), np.less(decDeg, bricksTab['DEC2']))
    centreMask=np.logical_and(RAMask, decMask)
    RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, 2*halfBoxSizeDeg)
    mask=np.logical_and(np.greater(bricksTab['RA1'], RAMin), np.less(bricksTab['RA2'], RAMax))
    if RAMin < 0:
        zeroCrossMask=np.greater(-(360-bricksTab['RA1']), RAMin)
        mask=np.logical_or(mask, zeroCrossMask)
    mask=np.logical_and(mask, np.greater(bricksTab['DEC1'], decMin))
    mask=np.logical_and(mask, np.less(bricksTab['DEC2'], decMax))
    mask=np.logical_or(mask, centreMask)
    matchTab=bricksTab[np.where(mask)]
    count=0
    tractorTabs=[]
    # try:
    #     matchTab=atpy.join(matchTab, DRTab, keys = 'BRICKNAME')
    # except:
    #     # Not in DECaLS?
    #     print("... no match between bricks and %s tables - %s ..." % (DR, outFileName))
    #     return None
    for mrow in matchTab:
        subDir="%03d" % np.floor(mrow['RA'])
        url=basePath+subDir
        url=url+os.path.sep+"tractor-"+mrow['BRICKNAME']+".fits"
        os.makedirs(cacheDir+os.path.sep+subDir, exist_ok = True)
        fileName=cacheDir+os.path.sep+subDir+os.path.sep+"tractor-%s.fits" % (mrow['BRICKNAME'])
        if os.path.exists(fileName) == False:
            print("... retrieving tractor catalog from web: %s ..." % (url))
            try:
                urllib.request.urlretrieve(url, filename = fileName)
            except:
                with open("wget_failed.sh", "a") as outFile:
                    outFile.write("wget -nc %s\n" % (url))
                print("... WARNING: failed to fetch from %s" % (url))
        try:
            print("... reading from cache: %s ..." % (fileName))
            tractorTab=atpy.Table.read(fileName)
            tractorTabs.append(tractorTab)
        except:
            print("... possibly a 404 error for %s - check if cached file is corrupted ..." % (fileName))
    if 'downloadOnly' in optionsDict.keys() and optionsDict['downloadOnly'] == True:
        return None
    
    # Stitch catalogs together
    if len(tractorTabs) > 0:
        tab=atpy.vstack(tractorTabs)
                
        # Cut to asked for size
        rDeg=astCoords.calcAngSepDeg(RADeg, decDeg, tab['ra'].data, tab['dec'].data)
        mask=np.less(rDeg, halfBoxSizeDeg)
        try:
            tab=tab[mask]
        except:
            raise Exception("Check if rDeg is not an array - %s, RADeg = %.6f, dec = %.6f" % (fileName, RADeg, decDeg))
        
        # DECaLS redshifts go very wrong when there are stars bright in W1, W2 in the vicinity
        # This should fix - we'll also throw out PSF-shaped sources too
        try:
            tab=tab[np.where(tab['brightblob'] == 0)]
        except:
            # Use maskbits 1, 11, 12, 13 for DR9
            # [still use brightblob for DR8 for now to be consistent with past results, but it's basically the same anyway]
            tab=tab[tab['maskbits'] != 2**1]
            tab=tab[tab['maskbits'] < 2**11]
        # Below is clean but loses us some clusters that look ok (we can't do proper bitwise anyway with atpy table?)
        #tab=tab[np.where(tab['maskbits'] == 0)]
        # Added getStars option for when we want to use this code for making
        if 'getStars' in optionsDict.keys() and optionsDict['getStars'] == True:
            starMask=np.logical_or(tab['type'] == 'PSF', tab['type'] == 'PSF ') # some catalogs have trailing space
            tab=tab[starMask]
            # Keep only Gaia objects - so we really, really get star-like objects
            tab=tab[tab['ref_cat'] == 'G2']
        else:
            tab=tab[np.where(tab['type'] != 'PSF')]
            tab=tab[np.where(tab['type'] != 'PSF ')] # Trailing space

        # WISE fluxes are available... i-band added in DECaLS DR10
        bands=['g', 'r', 'z', "w1", "w2"]# , 'Y']
        if DR == 'DR10':
            bands.append('i')
        
        # Convert nanomaggies to mags and do extinction correction
        bricksInTab=np.unique(tab['brickname'])
        for brickName in bricksInTab:
            # brickExtTab=DRTab[np.where(DRTab['BRICKNAME'] == brickName)]
            brickIndices=np.where(tab['brickname'] == brickName)
            brickMask=np.zeros(len(tab), dtype = bool)
            brickMask[brickIndices]=True
            for b in bands:
                magLabel='%s_mag' % (b)
                magErrLabel='%s_magErr' % (b)
                extKey='ext_%s' % (b)
                colsToAdd=[magLabel, magErrLabel]
                for col in colsToAdd:
                    if col not in list(tab.keys()):
                        tab.add_column(atpy.Column(np.ones(len(tab))*99., col))
                # Dust correction: correct flux upwards using mw_transmission (0-1)
                # Instead of using extinction mags below (seems to make no difference)
                nanoMaggies=np.array(tab['flux_%s' % (b)] / tab['mw_transmission_%s' % (b)])
                validMask=np.greater(nanoMaggies, 0.)
                mask=np.logical_and(brickMask, validMask)
                tab[magLabel][mask]=-2.5*np.log10(nanoMaggies[mask])+22.5
                validFluxErr=np.sqrt(1./tab['flux_ivar_%s' % (b)][mask])
                tab[magErrLabel][mask]=1./(nanoMaggies[mask]/validFluxErr)
                # Old...
                #if extKey in list(brickExtTab.keys()):
                    #dustCorrMag=brickExtTab[extKey][0]
                    #tab[magLabel][mask]=tab[magLabel][mask]-dustCorrMag
        
        catalog=[]
        for row in tab: 
            photDict={}
            photDict['id']=row['objid']
            photDict['RADeg']=row['ra']
            photDict['decDeg']=row['dec']
            for b in bands:
                photDict[b]=row['%s_mag' % (b)]
                photDict[b+"Err"]=row['%s_magErr' % (b)]
            if 'maxMagError' in list(optionsDict.keys()):
                keep=checkMagErrors(photDict, optionsDict['maxMagError'], bands = bands)
            else:
                keep=True
            if keep == True:
                catalog.append(photDict)
    
    else:
        catalog=None
    
    return catalog

#-------------------------------------------------------------------------------------------------------------
def DECaLSDR8Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, DR = None, optionsDict = {}):
    """Retrieves DECaLS DR8 tractor catalogs (if they exist) at the given position. Cuts the catalog to the
    radius specified by halfBoxSizeDeg.

    """
    
    stuff=DECaLSRetriever(RADeg, decDeg, halfBoxSizeDeg = halfBoxSizeDeg, DR = 'DR8', optionsDict = optionsDict)

    return stuff

#-------------------------------------------------------------------------------------------------------------
def DECaLSDR9Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, DR = None, optionsDict = {}):
    """Retrieves DECaLS DR9 tractor catalogs (if they exist) at the given position. Cuts the catalog to the
    radius specified by halfBoxSizeDeg.

    """
    
    stuff=DECaLSRetriever(RADeg, decDeg, halfBoxSizeDeg = halfBoxSizeDeg, DR = 'DR9', optionsDict = optionsDict)
    
    return stuff

#-------------------------------------------------------------------------------------------------------------
def DECaLSDR10Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, DR = None, optionsDict = {}):
    """Retrieves DECaLS DR10 tractor catalogs (if they exist) at the given position. Cuts the catalog to the
    radius specified by halfBoxSizeDeg.

    """

    stuff=DECaLSRetriever(RADeg, decDeg, halfBoxSizeDeg = halfBoxSizeDeg, DR = 'DR10', optionsDict = optionsDict)

    return stuff

#-------------------------------------------------------------------------------------------------------------
def DL_DECaLSDR10Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, DR = None, optionsDict = {}):
    """DECaLS DR10 retriever, using NOAO datalab.


    """

    makeCacheDir()
    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
    if os.path.exists(cacheDir) == False:
        os.makedirs(cacheDir, exist_ok = True)

    outFileName=cacheDir+os.path.sep+"DL_DECaLSDR10_%.4f_%.4f_%.2f.fits" % (RADeg, decDeg, halfBoxSizeDeg)
    if os.path.exists(outFileName) == False:
        RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, halfBoxSizeDeg)
        try:
            result=qc.query(sql='select objid, ra, dec, dered_mag_g, dered_mag_r, dered_mag_i, dered_mag_z, dered_mag_w1, dered_mag_w2,\
                                 flux_ivar_g, flux_ivar_r, flux_ivar_i, flux_ivar_z, flux_ivar_w1, flux_ivar_w2,\
                                 snr_g, snr_r, snr_i, snr_z, snr_w1, snr_w2, type, maskbits from ls_dr10.tractor where\
                                 RA BETWEEN %.6f AND %.6f AND DEC BETWEEN %.6f and %.6f' % (RAMin, RAMax, decMin, decMax),
                            fmt = 'table')
            result.write(outFileName, overwrite = True)
        except:
            result=None
            print("... WARNING: datalab query failed to get %s" % (outFileName))
    else:
        if 'fetchAndCacheOnly' in optionsDict.keys() and optionsDict['fetchAndCacheOnly'] == True:
            print("... already retrieved: %s ..." % (outFileName))
            return None
        print("... reading from cache: %s ..." % (outFileName))
        result=atpy.Table().read(outFileName)

    if result is None:
        return None

    # DECaLS redshifts go very wrong when there are stars bright in W1, W2 in the vicinity
    # This should fix - we'll also throw out PSF-shaped sources too
    tab=result
    if len(tab) == 0:
        return None
    tab=tab[tab['maskbits'] != 2**1]
    tab=tab[tab['maskbits'] < 2**11]
    tab=tab[np.where(tab['type'] != 'PSF')]
    tab=tab[np.where(tab['type'] != 'PSF ')] # Trailing space - probably not an issue on datalab

    # WISE fluxes are available... i-band added in DECaLS DR10
    bands=['g', 'r', 'i', 'z', "w1", "w2"]
    catalog=[]
    for row in tab:
        photDict={}
        photDict['id']=row['objid']
        photDict['RADeg']=row['ra']
        photDict['decDeg']=row['dec']
        # Photometric uncertainties now the same as regular DECaLS retriever
        for b in bands:
            if row['snr_%s' % (b)] > 0:
                photDict[b]=row['dered_mag_%s' % (b)]
                flux=np.power(10, (photDict[b]-22.5)/-2.5)  # nanomaggies
                fluxErr=np.sqrt(1./row['flux_ivar_%s' % (b)])
                photDict[b+'Err']=1./(flux/fluxErr)
            else:
                photDict[b]=99.0
                photDict[b+'Err']=99.0
        if 'maxMagError' in list(optionsDict.keys()):
            keep=checkMagErrors(photDict, optionsDict['maxMagError'], bands = bands)
        else:
            keep=True
        if keep == True:
            catalog.append(photDict)

    return catalog

#-------------------------------------------------------------------------------------------------------------
def RubinDP0Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Retrieve Rubin DP0 catalog data, assuming we have an API access token.

    """

    tap=optionsDict['TAP']
    strCenterCoords=str(RADeg)+", "+str(decDeg)
    strRadius=str(halfBoxSizeDeg)
    bands=['g', 'r', 'i', 'z']
    query="SELECT objectId, coord_ra, coord_dec "
    for band in bands:
        query=query+", %s_cModelFlux, %s_cModelFluxErr"% (band, band)
    query=query+" FROM dp02_dc2_catalogs.Object"
    query=query+" WHERE CONTAINS(POINT('ICRS', coord_ra, coord_dec), CIRCLE('ICRS', "+strCenterCoords+", "+strRadius+")) = 1 "
    query=query+" AND r_cModelFluxErr/r_cModelFlux > 5"
    query=query+" AND detect_isPrimary = 1 and r_extendedness = 1"
    tab=tap.run_sync(query).to_table()

    catalog=[]
    for row in tab:
        photDict={}
        photDict['id']=row['objectId']
        photDict['RADeg']=row['coord_ra']
        photDict['decDeg']=row['coord_dec']
        for b in bands:
            f=row['%s_cModelFlux' % (b)]
            ferr=row['%s_cModelFluxErr' % (b)]
            photDict[b]=99.0
            photDict[b+'Err']=99.0
            if ferr > 0:
                snr=f/ferr
                if snr > 0:
                    photDict[b]=-2.5*np.log10(f)+31.4
                    photDict[b+"Err"]=1/snr # do this properly later
        if 'maxMagError' in list(optionsDict.keys()):
            keep=checkMagErrors(photDict, optionsDict['maxMagError'], bands = bands)
        else:
            keep=True
        if keep == True:
            catalog.append(photDict)

    return catalog

#-------------------------------------------------------------------------------------------------------------
def DELVEDR2Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, DR = None, optionsDict = {}):
    """DELVE DR2 retriever, using NOAO datalab.

    """

    from dl import queryClient as qc
    RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, halfBoxSizeDeg)
    result=qc.query(sql='select ra, dec, mag_auto_g, mag_auto_r, mag_auto_i, mag_auto_z, magerr_auto_g, magerr_auto_r,\
                    magerr_auto_i, magerr_auto_z, extinction_g, extinction_r, extinction_i, extinction_z from delve_dr2.objects where\
                    extended_class_r > 1 and RA BETWEEN %.6f AND %.6f AND DEC BETWEEN %.6f and %.6f' % (RAMin, RAMax, decMin, decMax))

    catalog=[]
    count=0
    bands=['g', 'r', 'i', 'z']
    for row in result.split('\n')[1:]:
        count=count+1
        objDict={}
        objDict['id']=count
        bits=row.split(',')
        if len(bits) > 1:
            objDict['RADeg']=float(bits[0])
            objDict['decDeg']=float(bits[1])
            # Inc. extinction correction
            objDict['g']=float(bits[2])-float(bits[10])
            objDict['r']=float(bits[3])-float(bits[11])
            objDict['i']=float(bits[4])-float(bits[12])
            objDict['z']=float(bits[5])-float(bits[13])
            objDict['gErr']=float(bits[6])
            objDict['rErr']=float(bits[7])
            objDict['iErr']=float(bits[8])
            objDict['zErr']=float(bits[9])
            if 'maxMagError' in list(optionsDict.keys()):
                keep=checkMagErrors(objDict, optionsDict['maxMagError'], bands = bands)
            else:
                keep=True
            if keep == True:
                catalog.append(objDict)

    return catalog

#-------------------------------------------------------------------------------------------------------------
def SDSSDR7Retriever(RADeg, decDeg, halfBoxSizeDeg = 9.0/60.0, optionsDict = {}):
    """Retrieves SDSS DR7 main photometry at the given position.
    
    """
    
    stuff=SDSSRetriever(RADeg, decDeg, halfBoxSizeDeg = halfBoxSizeDeg, DR = 7, optionsDict = optionsDict)
    return stuff

#-------------------------------------------------------------------------------------------------------------
def SDSSDR8Retriever(RADeg, decDeg, halfBoxSizeDeg = 9.0/60.0, optionsDict = {}):
    """Retrieves SDSS DR8 photometry at the given position.
    
    """
    
    stuff=SDSSRetriever(RADeg, decDeg, halfBoxSizeDeg = halfBoxSizeDeg, DR = 8, optionsDict = optionsDict)
    return stuff

#-------------------------------------------------------------------------------------------------------------
def SDSSDR10Retriever(RADeg, decDeg, halfBoxSizeDeg = 9.0/60.0, optionsDict = {}):
    """Retrieves SDSS DR10 photometry at the given position.
    
    """
    
    stuff=SDSSRetriever(RADeg, decDeg, halfBoxSizeDeg = halfBoxSizeDeg, DR = 10, optionsDict = optionsDict)
    return stuff

#-------------------------------------------------------------------------------------------------------------
def SDSSDR12Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Retrieves SDSS DR12 photometry at the given position.
    
    """
    
    stuff=SDSSRetriever(RADeg, decDeg, halfBoxSizeDeg = halfBoxSizeDeg, DR = 12, optionsDict = optionsDict)
    return stuff

#-------------------------------------------------------------------------------------------------------------
def SDSSDR16Retriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Retrieves SDSS DR16 photometry at the given position.

    """

    stuff=SDSSRetriever(RADeg, decDeg, halfBoxSizeDeg = halfBoxSizeDeg, DR = 16, optionsDict = optionsDict)
    return stuff

#-------------------------------------------------------------------------------------------------------------
def CFHTLenSRetriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Retrieves CFHTLenS photometry, which works differently to other CFHT catalogues. 
    
    halfBoxSizeDeg is actually a radius in this case.
    
    NOTE: optionsDict must include 'maxMagError' key.
    
    """
    
    makeCacheDir()
    
    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
    
    # CFHTLens - note: mag 99s where missing, this does star/galaxy sep based on lensfit fitclass
    #url="http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/community/CFHTLens/cgi/queryt.pl?REQUEST=doQuery&LANG=ADQL&method=sync&format=ascii&query=SELECT%0D%0Aid%2C+ALPHA_J2000%2C+DELTA_J2000%2C+fitclass%2C+MAG_u%2C+MAGERR_u%2C+MAG_g%2C+MAGERR_g%2C+MAG_r%2C+MAGERR_r%2C+MAG_i%2C+MAGERR_i%2C+MAG_y%2C+MAGERR_y%2C+MAG_z%2C+MAGERR_z%0D%0AFROM%0D%0Acfht.clens%0D%0AWHERE%0D%0Afitclass%3E%3D0%0D%0AAND+fitclass%3C%3D0%0D%0AAND+contains%28pos%2Ccircle%28%27ICRS+GEOCENTER%27%2C$RADEG%2C$DECDEG%2C$SEARCHRAD%29%29%3D1%0D%0A" 
    # No star-galaxy separation
    #http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/cadcbin/community/CFHTLens/queryt.pl?REQUEST=doQuery&LANG=ADQL&method=sync&format=ascii&query=SELECT%0D%0Atop+10%0D%0Aid%0D%0AFROM%0D%0Acfht.clens%0D%0A
    # Updated March 2016: changed URLs at CADC
    url="http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/cadcbin/community/CFHTLens/queryt.pl?REQUEST=doQuery&LANG=ADQL&method=sync&format=ascii&query=SELECT%0D%0Aid%2C+ALPHA_J2000%2C+DELTA_J2000%2C+fitclass%2C+MAG_u%2C+MAGERR_u%2C+MAG_g%2C+MAGERR_g%2C+MAG_r%2C+MAGERR_r%2C+MAG_i%2C+MAGERR_i%2C+MAG_y%2C+MAGERR_y%2C+MAG_z%2C+MAGERR_z%0D%0AFROM%0D%0Acfht.clens%0D%0AWHERE%0D%0A+contains%28pos%2Ccircle%28%27ICRS+GEOCENTER%27%2C$RADEG%2C$DECDEG%2C$SEARCHRAD%29%29%3D1%0D%0A" 
    url=url.replace("$RADEG", "%.6f" % (RADeg))
    url=url.replace("$DECDEG", "%.6f" % (decDeg))
    url=url.replace("$SEARCHRAD", "%.6f" % (halfBoxSizeDeg))

    outFileName=cacheDir+os.path.sep+"CFHTLenS_%.4f_%.4f_%.4f.csv" % (RADeg, decDeg, halfBoxSizeDeg)

    print("... getting CFHTLenS photometry (file: %s) ..." % (outFileName)) 
        
    if os.path.exists(outFileName) == False:
        
        print("... fetching from the internet ...")
            
        response=None
        while response == None:
            try:
                response=urllib.request.urlopen(url)
            except:
                print("Network down, or CADC changed URLs again? Waiting 30 sec...")
                time.sleep(30)
                
        lines=response.read()
        lines=lines.split("\n")

        outFile=open(outFileName, "w")
        for line in lines:
            outFile.write(line+"\n")
        outFile.close()
    
    else:
        
        inFile=open(outFileName, "r")
        lines=inFile.readlines()
        inFile.close()

    # Parse .csv into catalog
    if len(lines) < 10:
        catalog=None
    else:
        catalog=[]
        idCount=0
        # NOTE: Extinction corrections already applied to CFHTLenS catalogs
        for line in lines[1:]: # first line always heading
            if len(line) > 3:
                photDict={}
                idCount=idCount+1
                bits=line.replace("\n", "").split()                
                photDict['id']=idCount    # just so we have something
                try:
                    photDict['RADeg']=float(bits[1])
                except:
                    print("... problem with file %s - removing and retrying in 70 sec ..." % (outFileName))
                    os.remove(outFileName)
                    time.sleep(70)
                # Note: skip mag y, old version of i-band filter
                photDict['decDeg']=float(bits[2])
                photDict['u']=float(bits[4])
                photDict['g']=float(bits[6])
                photDict['r']=float(bits[8])
                photDict['i']=float(bits[10])
                photDict['z']=float(bits[14])
                photDict['uErr']=float(bits[5])
                photDict['gErr']=float(bits[7])
                photDict['rErr']=float(bits[9])
                photDict['iErr']=float(bits[11])
                photDict['zErr']=float(bits[15])
                                                
                # Apply mag error cuts if given
                # We're just making the mag unconstrained here (missing data), rather than applying a limit
                # If we don't have a minimum of three useful bands, reject
                if 'maxMagError' in list(optionsDict.keys()):
                    keep=checkMagErrors(photDict, optionsDict['maxMagError'])
                else:
                    keep=True
                # If we want to keep what CFHTLenS classify as galaxies (galaxies are fitClass == 0)
                # NOTE: this makes no significant difference in our testing, so disabled
                #fitClass=int(bits[3])
                #if fitClass == 0:
                    #keep=True
                #else:
                    #keep=False
                if keep == True:
                    catalog.append(photDict)
                    
    return catalog

#-------------------------------------------------------------------------------------------------------------
def getEBMinusV(RADeg, decDeg, optionsDict = {}):
    """Get E(B-V) due to Galactic dust from Schlegel maps, using IRSA web service
    
    """

    makeCacheDir()
    if 'altCacheDir' in list(optionsDict.keys()):
        cacheDir=optionsDict['altCacheDir']
    else:
        cacheDir=CACHE_DIR
        
    fileName=cacheDir+os.path.sep+"SchlegelIRSA_%.6f_%.6f.xml" % (RADeg, decDeg)
    if os.path.exists(fileName) == False:
        url="http://irsa.ipac.caltech.edu/cgi-bin/DUST/nph-dust?locstr="+str(RADeg)+"+"+str(decDeg)+"+equ+J2000"
        urllib.request.urlretrieve(url, filename = fileName)
    
    inFile=open(fileName, "r")
    lines=inFile.readlines()
    inFile.close()
    for i in range(len(lines)):
        if lines[i].find("refPixelValueSFD") != -1:
            break
    try:
        EBMinusV=float(lines[i+1].split(" (")[0])
    except:
        raise Exception("failed to parse IRSA E(B-V) value - problem with file %s?" % (fileName))

    return EBMinusV

#-------------------------------------------------------------------------------------------------------------
def parseFITSPhotoTable(tab, fieldIDKey = None, optionsDict = {}):
    """Parse .fits table into catalog list of dictionaries format. Extinction correction due to Galactic
    dust will be applied, using IRSA web service. If fieldIDKey == None (the default), this is done
    at the mean RA, dec coords. Otherwise, sources are grouped by fieldID and the mean correction for
    each field is applied.
     
    Returns catalog (list of dictionaries)
    
    """

    # Options we may wish to play with
    magKey="MAG_AUTO"
    magErrKey="MAGERR_AUTO"
    magNumber=None
    #magKey="MAG_APER"
    #magErrKey="MAGERR_APER"
    #magNumber=2
        
    # Dust correction setup
    corrDict={'u': 5.155, 'g': 3.793, 'r': 2.751, 'i': 2.086, 'z': 1.479, 'Ks': 0.367}  
    EBMinusVList=[]
    if fieldIDKey == None:
        EBMinusV=getEBMinusV(np.mean(tab['RADeg']), np.mean(tab['decDeg']), optionsDict = optionsDict)
        EBMinusVList.append({'RADeg': np.mean(tab['RADeg']), 'decDeg': np.mean(tab['decDeg']), 'EBMinusV': EBMinusV})
    else:
        fieldNames=np.unique(tab['field'])
        for f in fieldNames:
            mask=np.where(tab['field'] == f)
            EBMinusV=getEBMinusV(np.mean(tab['RADeg'][mask]), np.mean(tab['decDeg'][mask]), optionsDict = optionsDict)
            EBMinusVList.append({'RADeg': np.mean(tab['RADeg'][mask]), 'decDeg': np.mean(tab['decDeg'][mask]), 'EBMinusV': EBMinusV})

    # Work out available bands
    acceptableBands=['u', 'g', 'r', 'i', 'z', 'Ks']
    tabBands=[]
    for key in list(tab.keys()):
        bits=key.split("_")
        if len(bits) > 0 and bits[0] in acceptableBands and bits[0] not in tabBands:
            tabBands.append(bits[0])
   
    # Parse into zCluster format, apply dust correction too
    catalog=[]
    for row in tab: 
        photDict={}
        photDict['id']=row['ID']
        photDict['RADeg']=row['RADeg']
        photDict['decDeg']=row['decDeg']
        rMin=1e6
        for EBMinusVDict in EBMinusVList:
            rDeg=astCoords.calcAngSepDeg(photDict['RADeg'], photDict['decDeg'], EBMinusVDict['RADeg'], EBMinusVDict['decDeg'])
            if rDeg < rMin:
                rMin=rDeg
                EBMinusV=EBMinusVDict['EBMinusV']
        for b in tabBands:
            dustCorrMag=EBMinusV*corrDict[b]
            if magNumber == None:
                photDict[b]=row['%s_%s' % (b, magKey)]-dustCorrMag
                photDict[b+"Err"]=row['%s_%s' % (b, magErrKey)]
            else:
                photDict[b]=row['%s_%s' % (b, magKey)][magNumber]-dustCorrMag
                photDict[b+"Err"]=row['%s_%s' % (b, magErrKey)][magNumber]
        catalog.append(photDict)

    return catalog

#------------------------------------------------------------------------------------------------------------
def addExtraPhoto(RADeg, decDeg, catalog, halfBoxSizeDeg = 9.0/60.0, optionsDict = {}):
    """Add extra photometry from another retriever, as specified in optionsDict using the 'extraRetriever'
    and 'extraOptions' keys. Used to supplement e.g. single band (Ks) catalogs from APO.
    
    """
    
    # Get extra catalog, if we want to include missing bands
    if 'extraRetriever' in list(optionsDict.keys()) and optionsDict['extraRetriever'] is not None: 
        extraRetriever=optionsDict["extraRetriever"]
        extraCat=extraRetriever(RADeg, decDeg, halfBoxSizeDeg = halfBoxSizeDeg, optionsDict = optionsDict['extraOptions'])
        if extraCat == None or extraCat == []:
            print("No objects found")
            return None

        RAs_extra=[]
        decs_extra=[]

        for objDict in extraCat:
            RAs_extra.append(objDict['RADeg'])
            decs_extra.append(objDict['decDeg'])
        RAs_extra=np.array(RAs_extra)
        decs_extra=np.array(decs_extra)
        crossMatchRadiusDeg=2.0/3600.0
        extraMatched=np.zeros(len(catalog))  # For sanity checking...
        count=0

        keysToSkip=['id', 'RADeg', 'decDeg']
        for objDict in catalog:
            rDeg=astCoords.calcAngSepDeg(objDict['RADeg'], objDict['decDeg'], RAs_extra, decs_extra)
            if rDeg.min() < crossMatchRadiusDeg:
                i=np.where(rDeg == rDeg.min())[0][0]
                extraObjDict=extraCat[i]
                for k in extraObjDict.keys():
                    if k not in list(keysToSkip) and k not in objDict.keys():
                        objDict['%s' % (k)]=extraObjDict[k]
                    # For sanity check below - we will do this properly somewhere else
                    #objDict['extra_%s' % (k)]=extraObjDict[k]
                    #objDict['extra_%sErr' % (k)]=extraObjDict["%sErr" % (k)]
                extraMatched[count]=1
            count=count+1

    if catalog == []:
        catalog=None

    return catalog

#-------------------------------------------------------------------------------------------------------------
def FITSRetriever(RADeg, decDeg, halfBoxSizeDeg = 36.0/60.0, optionsDict = {}):
    """Parses a FITS catalog made by e.g., soi_makecatalogs.py.
    
    NOTE: No star-galaxy separation is applied in this currently.
    
    Here optionsDict needs to include 'fileName' key.
    
    If 'addSDSS': True in optionsDict, then we fetch an SDSS catalog at the location, cross match it against 
    the FITS catalog, and add in info for bands we can't find (e.g., g-band). We don't add SDSS objects which
    aren't detected in the FITS catalog.

    """
    
    try:
        tab=atpy.Table().read(optionsDict['fileName'])
    except:
        raise Exception("assumed database is a FITS table file, but failed to read")
    
    # If the position isn't actually in our galaxy catalog, we give up now
    rDeg=astCoords.calcAngSepDeg(RADeg, decDeg, tab['RADeg'].data, tab['decDeg'].data)
    if rDeg.min() > halfBoxSizeDeg:
        print("... no galaxies found in FITS catalog near RA, dec = (%.6f, %.6f) ..." % (RADeg, decDeg))
        return None
    tab=tab[np.where(rDeg < halfBoxSizeDeg)]
    
    catalog=parseFITSPhotoTable(tab, optionsDict = optionsDict)
    catalog=addExtraPhoto(RADeg, decDeg, catalog, halfBoxSizeDeg = halfBoxSizeDeg, optionsDict = optionsDict)
        
    return catalog



"""

This module contains routines for estimating galaxy cluster photometric redshifts.

"""

import os
import sys
import numpy as np
from astLib import *
from scipy import interpolate
from scipy import stats
from scipy import ndimage
import astropy.io.fits as pyfits
from astropy.coordinates import SkyCoord
from astropy.coordinates import match_coordinates_sky
from PIL import Image
from PIL import ImageDraw
import pylab as plt

#------------------------------------------------------------------------------------------------------------
def getPixelAreaDeg2Map(mapData, wcs):
    """Returns a map of pixel area in square degrees.
    
    """
    
    # Get pixel size as function of position
    pixAreasDeg2=[]
    RACentre, decCentre=wcs.getCentreWCSCoords()
    x0, y0=wcs.wcs2pix(RACentre, decCentre)
    x1=x0+1
    for y0 in range(mapData.shape[0]):
        y1=y0+1
        ra0, dec0=wcs.pix2wcs(x0, y0)
        ra1, dec1=wcs.pix2wcs(x1, y1)
        xPixScale=astCoords.calcAngSepDeg(ra0, dec0, ra1, dec0)
        yPixScale=astCoords.calcAngSepDeg(ra0, dec0, ra0, dec1)
        pixAreasDeg2.append(xPixScale*yPixScale)
    pixAreasDeg2=np.array(pixAreasDeg2)
    pixAreasDeg2Map=np.array([pixAreasDeg2]*mapData.shape[1]).transpose()
    
    return pixAreasDeg2Map   

#---------------------------------------------------------------------------------------------------
def makeDegreesDistanceMap(degreesMap, wcs, RADeg, decDeg, maxDistDegrees):
    """Fills (in place) the 2d array degreesMap with distance in degrees from the given position,
    out to some user-specified maximum distance.

    Args:
        degreesMap (:obj:`np.ndarray`): Map (2d array) that will be filled with angular distance
            from the given coordinates. Probably you should feed in an array set to some extreme
            initial value (e.g., 1e6 everywhere) to make it easy to filter for pixels near the
            object coords afterwards.
        wcs (:obj:`astWCS.WCS`): WCS corresponding to degreesMap.
        RADeg (float): RA in decimal degrees of position of interest (e.g., object location).
        decDeg (float): Declination in decimal degrees of position of interest (e.g., object
            location).
        maxDistDegrees: The maximum radius out to which distance will be calculated.

    Returns:
        A map (2d array) of distance in degrees from the given position,
        (min x, max x) pixel coords corresponding to maxDistDegrees box,
        (min y, max y) pixel coords corresponding to maxDistDegrees box

    Note:
        This routine measures the pixel scale local to the given position, then assumes that it
        does not change. So, this routine may only be accurate close to the given position,
        depending upon the WCS projection used.

    """

    x0, y0=wcs.wcs2pix(RADeg, decDeg)
    ra0, dec0=RADeg, decDeg
    ra1, dec1=wcs.pix2wcs(x0+1, y0+1)
    xPixScale=astCoords.calcAngSepDeg(ra0, dec0, ra1, dec0)
    yPixScale=astCoords.calcAngSepDeg(ra0, dec0, ra0, dec1)

    xDistPix=int(round((maxDistDegrees)/xPixScale))
    yDistPix=int(round((maxDistDegrees)/yPixScale))

    Y=degreesMap.shape[0]
    X=degreesMap.shape[1]

    minX=int(round(x0))-xDistPix
    maxX=int(round(x0))+xDistPix
    minY=int(round(y0))-yDistPix
    maxY=int(round(y0))+yDistPix
    if minX < 0:
        minX=0
    if maxX > X:
        maxX=X
    if minY < 0:
        minY=0
    if maxY > Y:
        maxY=Y

    xDeg=(np.arange(degreesMap.shape[1])-x0)*xPixScale
    yDeg=(np.arange(degreesMap.shape[0])-y0)*yPixScale
    for i in range(minY, maxY):
        degreesMap[i][minX:maxX]=np.sqrt(yDeg[i]**2+xDeg[minX:maxX]**2)

    return degreesMap, [minX, maxX], [minY, maxY]

#------------------------------------------------------------------------------------------------------------
def makeBlankMap(RADeg, decDeg, sizePix, sizeDeg):
    """Makes a square blank map with a WCS.
    
    Returns:
        - mapData (2d array)
        - wcs (astWCS.WCS object)
    
    """
    CRVAL1, CRVAL2=RADeg, decDeg
    xSizeDeg, ySizeDeg=sizeDeg, sizeDeg
    xSizePix=float(sizePix)
    ySizePix=float(sizePix)
    xRefPix=xSizePix/2.0
    yRefPix=ySizePix/2.0
    xOutPixScale=xSizeDeg/xSizePix
    yOutPixScale=ySizeDeg/ySizePix
    newHead=pyfits.Header()
    newHead['NAXIS']=2
    newHead['NAXIS1']=xSizePix
    newHead['NAXIS2']=ySizePix
    newHead['CTYPE1']='RA---TAN'
    newHead['CTYPE2']='DEC--TAN'
    newHead['CRVAL1']=CRVAL1
    newHead['CRVAL2']=CRVAL2
    newHead['CRPIX1']=xRefPix+1
    newHead['CRPIX2']=yRefPix+1
    newHead['CDELT1']=-xOutPixScale
    newHead['CDELT2']=xOutPixScale    # Makes more sense to use same pix scale
    newHead['CUNIT1']='deg'
    newHead['CUNIT2']='deg'
    wcs=astWCS.WCS(newHead, mode='pyfits')
    
    return np.zeros([int(sizePix), int(sizePix)], dtype = float), wcs

#-------------------------------------------------------------------------------------------------------------
def makeDensityMap(RADeg, decDeg, catalog, z, dz = 0.1, rMaxMpc = 1.5, sizeMpc = 8.0, MpcPerPix = 0.1,
                   gaussSmoothPix = 2, outFITSFileName = None, outPlotFileName = None):
    """Makes a projected density map within +/- dz of the given redshift, smoothed using a Gaussian kernel,
    then calculates the centroid location, centroid shift (with respect to the original position), and an
    asymmetry statistic.
    
    Args:
        RADeg (float): RA coordinate (decimal degrees) of the center of the output map.
        decDeg (float): dec coordinate (decimal degrees) of the center of the output map.
        z (float): Central redshift at which the projected density will be calculated.
        dz (float, optional): Sets the integration range z +/- dz over which the projected density is
            calculated.
        rMaxMpc (float, optional): The maximum radial distance in Mpc (from RADeg, decDeg) in which the
            centroid search is performed.
        sizeMpc (float, optional): The size of the density map (which is square) in projected Mpc.
        MpcPerPix (float, optional): The pixel size, in projected Mpc, for the output density map.
        gaussSmoothPix (int, optional): The size of the Gaussian smoothing kernel applied to the output
            projected density map, in pixels.
        outFITSFileName (str, optional): If given, write the density map as a FITS file to the given
            location.
        makePlot (str, optional): If given, write a plot of the density map to the given location (the
            file format is determined by the extension, and supported formats depend on the matplotlib
            backend used).
        
    Returns:
        A dictionary with keys:
            - 'map': the output projected density map (2d array)
            - 'wcs': the output WCS
            - 'cRADeg': the RA of the centroid (peak) in the density map
            - 'cDecDeg': the dec of the centroid (peak) in the density map
            - 'offsetArcmin': size of the offset of the centroid (arcmin) from the original RA, dec
                coords
            - 'offsetMpc': size of the offset of the centroid (projected Mpc) from the original RA,
                dec coordinates
            - 'CS': centroid shift parameter (Mpc)
            - 'A': asymmetry parameter
    
    """
    
    # We go out to 4 Mpc for delta
    DA=astCalc.da(z)
    sizeDeg=np.degrees(sizeMpc/DA)
    sizePix=int(sizeMpc/MpcPerPix)
    d, wcs=makeBlankMap(RADeg, decDeg, sizePix, sizeDeg)
    for g in catalog:
        x, y=wcs.wcs2pix(g['RADeg'], g['decDeg'])
        x=int(round(x))
        y=int(round(y))
        mask=np.logical_and(g['pz_z'] > z-dz, g['pz_z'] < z+dz)
        v=np.trapz(g['pz'][mask], g['pz_z'][mask])
        if x >= 0 and x < d.shape[1] and y >= 0 and y < d.shape[0]:
            d[y, x]=d[y, x]+v
    d=ndimage.gaussian_filter(d, gaussSmoothPix)
    
    # Centre shift
    xpos,ypos=np.where(d==d.max())
    CS=np.sqrt((xpos-d.shape[1]/2)**2+(ypos-d.shape[0]/2)**2)[0]
    CSinMpc=CS*MpcPerPix

    # Asymmetry parameter
    d1=np.zeros(d.shape)+d
    d1=np.flipud(np.fliplr(d1))
    Adiff=np.power(d-d1, 2)
    A=np.sqrt((np.power(d-d1, 2).sum())/(2*np.power(d, 2).sum()))
    
    # We restrict centre finding to within 1.5 Mpc radius
    ys=np.array([np.arange(d.shape[0])]*d.shape[1]).transpose()
    xs=np.array([np.arange(d.shape[1])]*d.shape[0])
    coords=np.array(wcs.pix2wcs(xs.flatten(), ys.flatten()))
    ras=coords[:, 0]
    decs=coords[:, 1]
    rDegMap=astCoords.calcAngSepDeg(RADeg, decDeg, ras, decs).reshape(d.shape)
    rMpcMap=np.radians(rDegMap)*DA
    rMask=np.less(rMpcMap, rMaxMpc)
    y, x=np.where((d*rMask) == (d*rMask).max())
    cRADeg, cDecDeg=wcs.pix2wcs(x[0], y[0])
    offsetArcmin=astCoords.calcAngSepDeg(cRADeg, cDecDeg, RADeg, decDeg)*60
    offsetMpc=np.radians(offsetArcmin/60.)*DA
    
    if outFITSFileName is not None:
        astImages.saveFITS(outFITSFileName, d, wcs)

    if outPlotFileName is not None:
        axes=[0, 0, 1, 1]
        axesLabels="sexagesimal"    # Avoid dealing with axis flips
        figSize=(8.25, 8.25)
        fig=plt.figure(figsize = figSize, dpi = 300)
        p=astPlots.ImagePlot(d, wcs, cutLevels = [d.min(), d.max()],
                             colorMapName = 'magma', axes = axes,
                             axesLabels = axesLabels)

        if sizeMpc > 40:
            scaleBarMpc=5
        elif sizeMpc > 10 and sizeMpc < 40:
            scaleBarMpc=2
        elif sizeMpc < 10:
            scaleBarMpc=1
        scaleBarMpc=5#int(round(sizeMpc/5))
        scaleBarSizeArcmin=np.degrees(scaleBarMpc/astCalc.da(z))*60
        p.addScaleBar('SE', scaleBarSizeArcmin*60.0, color='cyan', fontSize=16, width=2.0, label = "%d Mpc" % (scaleBarMpc))
        plt.savefig(outPlotFileName)

    return {'map': d, 'wcs': wcs, 'cRADeg': cRADeg, 'cDecDeg': cDecDeg, 
            'offsetArcmin': offsetArcmin, 'offsetMpc': offsetMpc, 'CS': CS, 'A': A}

#-------------------------------------------------------------------------------------------------------------
def makeWeightedNz(RADeg, decDeg, catalog, zPriorMax, weightsType, minDistanceMpc = 0.0, maxDistanceMpc = 1.0, 
                   applySanityCheckRadius = True, sanityCheckRadiusArcmin = 1.0, areaMask = None, wcs = None):
    """Make a N(z) distribution in the direction of the postion (usually of a cluster) given by RADeg, decDeg.
    This is constructed from the p(z) distributions of all galaxies in the catalog, with radial weights applied
    as per weightsType. So, for 'flat1Mpc', the radial weight is 1 if within a projected distance of 1 Mpc of 
    RADeg, decDeg, and 0 otherwise.
    
    zPriorMax can be used to reign in any spuriously high redshifts given our prior knowledge of the depth of a
    given photometric survey. However, this is now largely redundant as the mag-based prior does most of the
    work (see PhotoRedshiftEngine.py).
    
    NOTE: minDistanceMpc only applies to 'flat' weightsType (for local background estimate)
    
    """

    # For speed, we make a big array of all p(z) here
    pzArray=[]
    for obj in catalog:
        if 'pz' in list(obj.keys()):
            pzArray.append(obj['pz'])
    pzArray=np.array(pzArray)
    
    # Make big array of angular separation for each galaxy
    # Get EAZY odds too, in case useful for weighting
    gOdds, gRedshifts, angSepArray, tanArray=extractArraysFromGalaxyCatalog(catalog, RADeg, decDeg)
    
    # Sanity check: if we didn't find a decent number of galaxies close to the cluster centre, 
    # the cluster is probably not actually in the optical catalog footprint (e.g., just outside edge of S82)
    if applySanityCheckRadius == True and np.sum(np.less(np.degrees(tanArray), sanityCheckRadiusArcmin/60.0)) == 0:
        print("... no galaxies within %.1f' of cluster position ..." % (sanityCheckRadiusArcmin))
        return None

    # Assume same z grid for all objects, calculate DAs
    for obj in catalog:
        if 'pz_z' in list(obj.keys()):
            zArray=obj['pz_z']
            break
    DAArray=[]
    for z in zArray:
        DAArray.append(astCalc.da(z))
    DAArray=np.array(DAArray)
        
    # NOTE: radial weights should have max value 1, min value 0
    # Koester, Postman et al. style radial weights (projected 2d NFW profile)
    if weightsType == 'NFW':
        rs=0.15 # rs=R200/c c=concentration - Koester uses 150 kpc
        r=np.linspace(0.0, 10.0, 2000)
        x=r/rs
        fx=np.zeros(x.shape)
        mask=np.greater(x, 1)
        fx[mask]=1-(2.0/np.sqrt(x[mask]**2-1))*np.arctan(np.sqrt((x[mask]-1)/(x[mask]+1)))
        mask=np.logical_and(np.greater(x, 0), np.less(x, 1))
        fx[mask]=1-(2.0/np.sqrt(1-x[mask]**2))*np.arctanh(np.sqrt((1-x[mask])/(x[mask]+1)))
        mask=np.equal(x, 1)
        fx[mask]=0
        mask=np.greater(x, 20)
        fx[mask]=0
        sigmax=np.zeros(r.shape)
        mask=np.greater(r, rs)
        sigmax[mask]=(2*rs*fx[mask])/(x[mask]**2-1)
        mask=np.less(r, rs)
        sigmax[mask]=sigmax.max()
        sigmax=sigmax/sigmax.max()
        tckSigmax=interpolate.splrep(r, sigmax)
        rMpc=(np.array([tanArray]*DAArray.shape[0]).transpose())*DAArray
        rWeights=[]
        for row in rMpc:
            rWeights.append(interpolate.splev(row, tckSigmax))
        rWeights=np.array(rWeights)
        rWeights[np.less(rWeights, 0)]=0.0
        # We have to match area to background: so, zero weight all galaxies outside 1 Mpc radius
        rWeights[np.greater(rMpc, maxDistanceMpc)]=0.0
    
    elif weightsType == 'flat':
    
        # Calculate weights - set to zero for objects outside of fiducial projected distance
        # 1 inside 1 Mpc radius, 0 otherwise
        # Use this for background (the 1 Mpc radius just normalises area
        rMpc=(np.array([tanArray]*DAArray.shape[0]).transpose())*DAArray
        rWeights=np.ones(rMpc.shape)
        rWeights[np.greater(rMpc, maxDistanceMpc)]=0.0
        rWeights[np.less(rMpc, minDistanceMpc)]=0.0
        
    elif weightsType == 'radial':
        
        # Calculate weights - set to zero for objects outside of fiducial projected distance
        # 1 if at centre coords, falls linearly with increasing radius
        rMpc=(np.array([tanArray]*DAArray.shape[0]).transpose())*DAArray
        rWeights=1.0-rMpc/maxDistanceMpc
        rWeights[np.greater(rMpc, maxDistanceMpc)]=0.0
    
    else:
        raise Exception("didn't understand weightsType")
    
    # Simplest odds weighting scheme is to cut everything with odds < some value
    # Whatever happens, we definitely want to cut anything with odds == 0, as its z will default to zMin
    # NOTE: leave set at gOddsCut=0.5 - this fixed null test problems and didn't negatively impact z accuracy
    gOddsCut=0.5
    gOddsWeights=np.ones(gOdds.shape)
    mask=np.less(gOdds, gOddsCut)
    gOddsWeights[mask]=0.0    
    gOddsWeights=[gOddsWeights]*rWeights.shape[1]
    gOddsWeights=np.array(gOddsWeights).transpose()
    #gOddsWeights[:]=1 # uncomment to disable gOdds weights
    
    # NOTE: zPrior applied in calculating cluster redshift later, not in counting galaxies here
    Nz_r=rWeights.sum(axis = 0)                                             # Number of galaxies within 1 Mpc at each z (IF all weights were 1)
    Nz_r_odds_total=np.greater(rWeights*gOddsWeights, 0).sum(axis = 0)      # Actual number of galaxies within 1Mpc at each z regardless of weight (if using NFW, radial - for flat, Nz_total == Nz_r)
    Nz_r_odds=np.sum(rWeights*gOddsWeights, axis = 0)   # Number of galaxies within 1 Mpc at each z AND odds > cut
    NzWeightedSum=np.sum(pzArray*rWeights*gOddsWeights, axis = 0)
    
    # We may as well keep track of area as well
    if areaMask is None:
        areaMpc2=np.array([np.pi*maxDistanceMpc**2 - np.pi*minDistanceMpc**2]*len(zArray))
    else:
        areaMap=getPixelAreaDeg2Map(areaMask, wcs)
        rDegMap=np.zeros(areaMask.shape)
        rDegMap, xRange, yRange=makeDegreesDistanceMap(rDegMap, wcs, RADeg, decDeg, 2.0)
        rDegMap=rDegMap[yRange[0]:yRange[1], xRange[0]:xRange[1]]
        areaMap=areaMask[yRange[0]:yRange[1], xRange[0]:xRange[1]]*areaMap[yRange[0]:yRange[1], xRange[0]:xRange[1]]
        areaMpc2=[]
        count=0
        for DA in DAArray:
            areaScaling=1/np.power(np.degrees(np.arctan(1.0/DA)), 2)
            max_rDegCut=np.degrees(np.arctan(maxDistanceMpc/DA))
            min_rDegCut=np.degrees(np.arctan(minDistanceMpc/DA))
            areaDeg2=areaMap[np.logical_and(rDegMap >= min_rDegCut, rDegMap < max_rDegCut)].sum()
            areaMpc2.append(areaScaling*areaDeg2)
            count=count+1
        areaMpc2=np.array(areaMpc2)

    # If we want p(z), we can divide pzWeightedSum by Nz_r_odds
    return {'NzWeightedSum': NzWeightedSum, 'Nz': Nz_r_odds, 'Nz_total': Nz_r_odds_total, 'zArray': zArray, 
            'DAArray': DAArray, 'areaMpc2': areaMpc2}
    
#-------------------------------------------------------------------------------------------------------------
def estimateClusterRedshift(RADeg, decDeg, catalog, zPriorMin, zPriorMax, weightsType, maxRMpc, 
                            zMethod, maskMap = None, maskWCS = None, sanityCheckRadiusArcmin = 1.0, 
                            bckCatalog = [], bckAreaDeg2 = None, filterDeltaValues = True,
                            minBackgroundAreaMpc2 = 11.0):
    """This does the actual work of estimating cluster photo-z from catalog.
    
    Assumes each object has keys 'pz' (p(z), probability distribution), 'pz_z' (corresponding redshifts at 
    each point in p(z)
    
    If bckCatalog and bckAreaDeg2 are given, then these are used for the background estimation. If not, then
    a projected 3-4 Mpc annulus is used.
    
    If zDebias is not None, the final output redshift ('z') will be:
        
        z_final = z_initial + zDebias*(1+z_initial) 
    
    This is obviously a fudge, so use with caution (only used for some surveys - see bin/zCluster).
            
    """
        
    # Initial sanity check: if we didn't find a decent number of galaxies close to the cluster centre, 
    # the cluster is probably not actually in the optical catalog footprint (e.g., just outside edge of S82)
    gOdds, gRedshifts, angSepArray, tanArray=extractArraysFromGalaxyCatalog(catalog, RADeg, decDeg)
    if np.sum(np.less(np.degrees(tanArray), sanityCheckRadiusArcmin/60.0)) == 0:
        return None

    # Automated area mask calculation, using only catalogs
    # Probably not super accurate but better than nothing - will at least take care of survey boundaries
    if maskMap is None and maskWCS is None:
        areaMask, wcs=estimateAreaMask(RADeg, decDeg, catalog)
    else:
        areaMask, wcs=maskMap, maskWCS
    
    # Here, we do a weighted average of the p(z) distribution
    # The weighting is done according to distance from the cluster position at the z corresponding to a
    # given p(z). 
    clusterNzDict=makeWeightedNz(RADeg, decDeg, catalog, zPriorMax, weightsType, maxDistanceMpc = 1.0,
                                 areaMask = areaMask, wcs = wcs)
    zArray=clusterNzDict['zArray']
    if clusterNzDict['NzWeightedSum'].sum() == 0:
        return None
    
    # This is how we have been calculating redshifts
    # We only need normFactor to make sure odds is correct
    pzWeightedMean=clusterNzDict['NzWeightedSum']/clusterNzDict['Nz']
    pzWeightedMean[np.isnan(pzWeightedMean)]=0.0
    normFactor=np.trapz(pzWeightedMean, clusterNzDict['zArray'])
    pzWeightedMean=pzWeightedMean/normFactor
    z, odds, zOdds=calculateRedshiftAndOdds(pzWeightedMean, clusterNzDict['zArray'], dzOdds = 0.05, method = zMethod, 
                                            zPriorMax = zPriorMax, zPriorMin = zPriorMin)

    # Background used in delta (density contrast) measurement
    # We also re-do the N(z) measurement in the cluster region so that this makes sense
    clusterNzDictForSNR=makeWeightedNz(RADeg, decDeg, catalog, zPriorMax, 'flat', maxDistanceMpc = maxRMpc)
    if len(bckCatalog) > 0:
        # Global
        bckNzDict=makeWeightedNz(RADeg, decDeg, bckCatalog, zPriorMax, 'flat', minDistanceMpc = 3.0, maxDistanceMpc = 1e9, 
                                 applySanityCheckRadius = False)
        bckNzDict['areaMpc2']=((np.radians(np.sqrt(bckAreaDeg2))*bckNzDict['DAArray'])**2)
    else:
        # Local background - 0.6 deg radius gives us out to 4 Mpc radius at z = 0.1
        bckNzDict=makeWeightedNz(RADeg, decDeg, catalog, zPriorMax, 'flat', minDistanceMpc = 3.0, maxDistanceMpc = 4.0,
                                 areaMask = areaMask, wcs = wcs)
    # Masking sanity check that isn't fully implemented (in some cases would just need to bump up mask resolution)
    if max(bckNzDict['areaMpc2']) > (np.pi*(4**2-3**2)): 
        print("... WARNING: max(areaMpc2) = %.3f > %.3f at (RADeg, decDeg) = (%.6f, %.6f)" % (max(bckNzDict['areaMpc2']), np.pi*(4**2-3**2), RADeg, decDeg))        
        #raise Exception("Mask resolution is too coarse - area from mask > area of circular annulus") 
    
    # Delta calculation with bootstrap error over the whole z range
    # For the first part of the mask, we allow background area == 1/2 of the 3-4 Mpc ring (previous min value was 20)
    # This allows us to get out to survey boundaries but will have large delta error bars
    validMask=np.greater_equal(bckNzDict['areaMpc2'], minBackgroundAreaMpc2) 
    bckAreaNorm=np.zeros(len(clusterNzDictForSNR['areaMpc2']))
    bckAreaNorm[validMask]=clusterNzDictForSNR['areaMpc2'][validMask]/bckNzDict['areaMpc2'][validMask]
    validMask=np.logical_and(validMask, clusterNzDictForSNR['NzWeightedSum'] > 0)
    validMask=np.logical_and(validMask, bckNzDict['NzWeightedSum'] > 0)
    if validMask.sum() > 0:
        nc=clusterNzDictForSNR['NzWeightedSum'][validMask]
        nb=bckNzDict['NzWeightedSum'][validMask]
        err_nc=np.sqrt(nc)
        err_nb=np.sqrt(nb)
        A=bckAreaNorm[validMask]
        delta=(nc/(A*nb))-1
        bs_delta=[]
        for i in range(5000):   # Needed for this to converge at 2 decimal places after rounding
            bs_nc=np.random.poisson(nc)
            bs_nb=np.random.poisson(nb)
            bs_nb[bs_nb == 0]=1 # Just to avoid div 0 warnings... background should never be 0 anyway
            bs_delta.append((bs_nc/(A*bs_nb))-1)
        errDelta=np.std(bs_delta, axis = 0)
        errDelta=np.round(errDelta, 2)
        
        # This uses the old delta method (effectively)
        zDelta=zArray[validMask]
        try:
            deltaIndex=np.where(zDelta == z)[0][0]
            delta_at_z=delta[deltaIndex]
            errDelta_at_z=errDelta[deltaIndex]
        except:
            print("... no valid delta value ...")
            return None
        
        # Optionally filter zOdds according to whether delta is > 3 sigma
        if filterDeltaValues == True:
            if zMethod == 'odds':
                pzToCheck=zOdds
            elif zMethod == 'max':
                pzToCheck=pzWeightedMean
            pzToCheck=applyUniformPrior(pzToCheck, zArray, zPriorMax = zPriorMax, zPriorMin = zPriorMin) 
            pzToCheck_validDelta=np.greater(delta/errDelta, 3)*pzToCheck[validMask]
            zIndex=np.argmax(pzToCheck_validDelta)
            z=zDelta[zIndex]
            delta_at_z=delta[zIndex]
            errDelta_at_z=errDelta[zIndex]
            if z < zPriorMin or z > zPriorMax:
                print("... not within z prior range ...")
                return None
        assert(np.isinf(delta_at_z) == False)
            
    else:
        print("... background area too small - skipping ...")
        return None
    
    return {'z': z, 'pz': pzWeightedMean, 'zOdds': zOdds, 'pz_z': zArray, 'delta': delta_at_z, 'errDelta': errDelta_at_z,
            'areaMask': areaMask, 'wcs': wcs}

#-------------------------------------------------------------------------------------------------------------
def estimateAreaMask(RADeg, decDeg, catalog):
    """Using the objects in the catalog, estimate the area covered - to take care of e.g. survey boundaries
    without needing to get masks for every survey. This works by estimating the average angular separation 
    between sources in the catalog, and using that to draw a circle of that radius around every object. This 
    is then used as the mask in area calculations.
    
    Args:
        RADeg (float): Cluster RA position in decimal degrees.
        decDeg (float): Cluster dec position in decimal degrees.
        catalog (list): List of dictionaries, where each dictionary defines an object in the catalog.
    
    Returns:
        areaMask (2d array): Mask of area covered.
        wcs (astWCS.WCS object): WCS corresponding to areaMask.
    
    """

    # Typical separation in degrees - use to set scale
    RAs=[]
    decs=[]
    for obj in catalog:
        RAs.append(obj['RADeg'])
        decs.append(obj['decDeg'])
    RAs=np.array(RAs)
    decs=np.array(decs)
    catCoords=SkyCoord(ra = RAs, dec = decs, unit = 'deg')
    xIndices, rDeg, sep3d = match_coordinates_sky(catCoords, catCoords, nthneighbor = 2)
    sepDeg=np.median(rDeg.data)
    
    # Make mask - max radius here is from catalogRetriever but should be adjustable elsewhere
    maxRadiusDeg=40/60.0
    widthPix=int(np.ceil((2*maxRadiusDeg)/sepDeg))
    areaMask=np.zeros([widthPix, widthPix])
    im=Image.fromarray(areaMask)
    draw=ImageDraw.Draw(im)
    CRVAL1, CRVAL2=RADeg, decDeg
    xSizeDeg, ySizeDeg=2*maxRadiusDeg, 2*maxRadiusDeg
    xSizePix=float(widthPix)
    ySizePix=float(widthPix)
    xRefPix=xSizePix/2.0
    yRefPix=ySizePix/2.0
    xOutPixScale=xSizeDeg/xSizePix
    yOutPixScale=ySizeDeg/ySizePix
    newHead=pyfits.Header()
    newHead['NAXIS']=2
    newHead['NAXIS1']=xSizePix
    newHead['NAXIS2']=ySizePix
    newHead['CTYPE1']='RA---TAN'
    newHead['CTYPE2']='DEC--TAN'
    newHead['CRVAL1']=CRVAL1
    newHead['CRVAL2']=CRVAL2
    newHead['CRPIX1']=xRefPix+1
    newHead['CRPIX2']=yRefPix+1
    newHead['CDELT1']=-xOutPixScale
    newHead['CDELT2']=xOutPixScale
    newHead['CUNIT1']='deg'
    newHead['CUNIT2']='deg'
    wcs=astWCS.WCS(newHead, mode='pyfits')
    for obj in catalog:
        x, y=wcs.wcs2pix(obj['RADeg'], obj['decDeg'])
        x=int(x); y=int(y)
        #areaMask[y, x]=1
        draw.ellipse([x-2, y-2, x+2, y+2])
    areaMask=np.array(im)

    return areaMask, wcs
    
#-------------------------------------------------------------------------------------------------------------
def extractArraysFromGalaxyCatalog(catalog, RADeg, decDeg):
    """For convenience: pulls out arrays needed for NGal measurement from galaxy catalog. Needs RADeg, decDeg
    to calculate angular distances from cluster (or random) position.
    
    """
    
    RAs=[]
    decs=[]
    gOdds=[]
    gRedshifts=[]
    #gTemplates=[]
    for obj in catalog:
        RAs.append(obj['RADeg'])
        decs.append(obj['decDeg'])
        if 'pz' in list(obj.keys()):
            gOdds.append(obj['odds'])
            gRedshifts.append(obj['zPhot'])
            #gTemplates.append(obj['bestFitSEDFileName'])
    RAs=np.array(RAs)
    decs=np.array(decs)
    angSepArray=astCoords.calcAngSepDeg(RADeg, decDeg, RAs, decs)
    tanArray=np.radians(angSepArray)
    gOdds=np.array(gOdds)
    gRedshifts=np.array(gRedshifts)
    
    return gOdds, gRedshifts, angSepArray, tanArray

#-------------------------------------------------------------------------------------------------------------
def applyUniformPrior(pz, zArray, zPriorMin = None, zPriorMax = None):
    """Applies uniform redshift prior to pz between zPriorMin and zPriorMax.
    
    """
    
    prior=np.ones(pz.shape)
    if zPriorMax is not None:
        prior[np.greater(zArray, zPriorMax)]=0.0
    if zPriorMin is not None:
        prior[np.less(zArray, zPriorMin)]=0.0
    pz=pz*prior
    norm=np.trapz(pz, zArray)
    pz=pz/norm
    
    return pz
    
#-------------------------------------------------------------------------------------------------------------
def calculateRedshiftAndOdds(pz, zArray, dzOdds = 0.2, method = 'max', zPriorMax = None, zPriorMin = None):
    """Calculates z and BPZ/EAZY style odds for given pz, zArray
    
    Method == 'max' finds z based on max peak in pz
    Method == 'odds' finds z based on z at which odds is maximised
    
    """
        
    # Prior: to avoid worst cases of overestimating z
    pz=applyUniformPrior(pz, zArray, zPriorMax = zPriorMax, zPriorMin = zPriorMin) 
        
    zToIndex_tck=interpolate.splrep(zArray, np.arange(zArray.shape[0]))
    z=zArray[pz.tolist().index(pz.max())]
    
    zOdds=None  # we only calculate this if method == 'odds', but we return always
    
    if method == 'max':
        zMin=z-dzOdds
        zMax=z+dzOdds
        indexMin, indexMax=interpolate.splev([zMin, zMax], zToIndex_tck)
        indexMin=int(round(indexMin))
        indexMax=int(round(indexMax))
        if indexMin < 0:
            indexMin=0
        if indexMax > pz.shape[0]-1:
            indexMax=pz.shape[0]-1
        odds=np.trapz(pz[indexMin:indexMax], zArray[indexMin:indexMax])
        
    elif method == 'odds':
        zOdds=[]
        for z in zArray:
            zMin=z-dzOdds
            zMax=z+dzOdds
            indexMin, indexMax=interpolate.splev([zMin, zMax], zToIndex_tck)
            indexMin=int(round(indexMin))
            indexMax=int(round(indexMax))
            if indexMin < 0:
                indexMin=0
            if indexMax > pz.shape[0]-1:
                indexMax=pz.shape[0]-1
            odds=np.trapz(pz[indexMin:indexMax], zArray[indexMin:indexMax])
            zOdds.append(odds)
        zOdds=np.array(zOdds)
        z=zArray[np.argmax(zOdds)]
        odds=zOdds[np.argmax(zOdds)]
        #--
        norm=np.trapz(zOdds, zArray)
        zOdds=zOdds/norm
            
    return [z, odds, zOdds]
    



"""

This module contains routines related to handling stellar population models,
for the purpose of stellar mass estimates.

"""

import os
import sys
import numpy as np
import astropy.table as atpy
import glob
import time
from astLib import *
from scipy import interpolate
from scipy import stats
import pylab
import itertools

#-------------------------------------------------------------------------------------------------------------
def setUpStellarMassSEDs(stellarMassModelDir, passbandsList, z, solarMetallicityOnly = True):
    """Set-up stellar population synthesis model SEDs for the purposes of stellar mass estimation.

    Only BC03 format is supported at the moment.

    """

    print(">>> Setting up stellar population models")
    t0=time.time()
    EBMinusVList=np.linspace(0, 0.48, 13).tolist()
    modelSEDDictList=[]
    if solarMetallicityOnly == True:
        modelSEDFileNames=glob.glob("%s/bc03_m62*.gplm" % (stellarMassModelDir))
    else:
        modelSEDFileNames=glob.glob("%s/bc03_*.gplm" % (stellarMassModelDir))
    count=0
    for modelSEDFileName in modelSEDFileNames:
        count=count+1
        print("... %d/%d ..." % (count, len(modelSEDFileNames)))
        modelSED=astSED.BC03Model(modelSEDFileName)
        # Only use ages less than age of the Universe
        ages=np.array(modelSED.ages)[np.less(modelSED.ages, astCalc.tz(z))]
        # Only load the mass file once per model SED - this is a bottle neck if we repeat for every age!
        modelMassDict=loadBC03MassFile(modelSEDFileName)
        #t00=time.time()
        for ageGyr in ages:
            s=modelSED.getSED(ageGyr, z = z)
            for EBMinusV in EBMinusVList:
                if EBMinusV > 0:
                    s.extinctionCalzetti(EBMinusV) # modifies z0flux, so this should be okay
                modelSEDDict=s.getSEDDict(passbandsList)
                modelSEDDict['labels']="label"
                modelSEDDict['EBMinusV']=EBMinusV
                modelSEDDict['ageGyr']=ageGyr
                modelSEDDict['z']=z
                modelSEDDict['fileName']=modelSED.fileName
                modelSEDDict['metallicity']=getMetallicity("BC03", modelSED.fileName)
                modelSEDDict['tauGyr']=getTauGyr("BC03", modelSED.fileName)
                modelSEDDict['modelListIndex']=count
                modelSEDDict['stellarMass']=interpolate.splev(modelSEDDict['ageGyr'], modelMassDict['tckMass'])
                modelSEDDictList.append(modelSEDDict)
    t1=time.time()
    print("... took %.3f sec" % (t1-t0))

    return modelSEDDictList

#-------------------------------------------------------------------------------------------------------------
def loadBC03MassFile(modelPath):
    """Loads BC03 *.4color files, which contain the stellar mass info, corresponding to the given model path.
    Returns dictionaries with numpy arrays of the different masses as a fn. of age., plus output from
    splrep which can be used to interpolate mass from age - e.g. interpolate.splev([ageGyr], dict['tckMass']

    Also extract tau from modelPath and use to calculate SFR evolution, returned in keys 'tauGyr', 'SFR'.
    Multiply SFR by stellar mass to get SFR in solar masses per year.

    Heck, while we're at it, parse the metallicity from modelPath into something sensible too.

    """

    massPath=modelPath.replace(".gplm", ".4color")
    inFile=open(massPath, "r")
    lines=inFile.readlines()
    inFile.close()
    age=[]
    Mstar=[]
    Mbol=[]
    MStarToLV=[]
    Vabs=[]    # abs mag in rest frame V through Buser V filter
    for line in lines:
        if line[0] != "#" and len(line) > 3:
            bits=line.split()
            if len(bits) > 3:   # incomplete lines for some CB07 files - not sure if this is a problem
                age.append((10.0**float(bits[0]))/1e9)
                Mstar.append(float(bits[6]))
                Mbol.append(float(bits[1]))
                MStarToLV.append(float(bits[5]))
                Vabs.append(float(bits[3]))
            #else:
                #print "What now?"
                #ipshell()
                #sys.exit()
    age=np.array(age)
    Mstar=np.array(Mstar)
    Mbol=np.array(Mbol)

    # Sanity check - does the mass file contain a similar range of ages to the model file?
    # If not, masses can be garbage as the interpolation gets extrapolated in a stupid fashion
    m=astSED.BC03Model(modelPath)
    if max(m.ages) > age.max():
        raise(Exception, "Mass file corresponding to %s is truncated (not enough ages)" % (modelPath))

    # All of the cubic interpolation here seems somewhat flakey for BC03/CB07?
    s=interpolate.splrep(age, Mstar, k=1)
    splMbol=interpolate.splrep(age, Mbol, k=1)
    splVabs=interpolate.splrep(age, Vabs, k=1)
    splMStarToLV=interpolate.splrep(age, MStarToLV, k=1)

    # Parse tau for tau models from file name according to my convention, set up interpolation from age too
    tauGyr=float(modelPath[modelPath.find("tau")+3:modelPath.find(".gplm")].replace("p", "."))
    SFR=(1.0/tauGyr)*np.exp(-age/tauGyr)/1e9
    sSFR=interpolate.splrep(age, SFR, k=1)

    # Parse metallicity into ZSun value - these from BC03 README file
    labels=["m22", "m32", "m42", "m52", "m62", "m72"]
    metallicities=[0.005, 0.02, 0.2, 0.4, 1.0, 2.5]
    labelInFileName=modelPath[modelPath.find("_m")+1:modelPath.find("_m")+4]
    ZSun=metallicities[labels.index(labelInFileName)]

    return {'ageGyr': age, 'Mstar': Mstar, 'ZSun': ZSun, 'tauGyr': tauGyr, 'tckMass': s, 'tckSFR': sSFR,
            'Mbol': Mbol, 'Vabs': Vabs, 'tckMbol': splMbol, 'tckVabs': splVabs, 'tckMStarToLV': splMStarToLV}

#-------------------------------------------------------------------------------------------------------------
def getTauGyr(modelFamily, modelPath):
    """Returns value of tau (SFH) by parsing the modelFileName

    """
    if modelFamily == "M05":
        tauGyr=float(modelPath[modelPath.find("_e")+2:modelPath.find("_z")])
    elif modelFamily == "BC03":
        tauGyr=float(modelPath[modelPath.find("tau")+3:modelPath.find(".gplm")].replace("p", "."))
    else:
        raise(Exception, "Need to add code to parse tau values from model file names for this family")

    return tauGyr

#-------------------------------------------------------------------------------------------------------------
def getMetallicity(modelFamily, modelPath):
    """Returns value of tau (SFH) by parsing the modelFileName

    """
    if modelFamily == "M05":
        print("no code to get M05 metallicity yet")
        ipshell()
        sys.exit()
        tauGyr=float(modelPath[modelPath.find("_e")+2:modelPath.find("_z")])
    elif modelFamily == "BC03":
        # From BC03 README
        ZSun=0.02
        ZMapDict={'m22': 0.0001/ZSun,
                  'm32': 0.0004/ZSun,
                  'm42': 0.004/ZSun,
                  'm52': 0.008/ZSun,
                  'm62': 0.02/ZSun,
                  'm72': 0.05/ZSun}
        fileNameKey=os.path.split(modelPath)[-1].split("_")[1]
        Z=ZMapDict[fileNameKey]
    else:
        raise(Exception, "Need to add code to parse tau values from model file names for this family")

    return Z

#------------------------------------------------------------------------------------------------------------
def fitSEDDictAndCalcStellarMass(SEDDict, modelSEDDictList, distNorm):
    """Custom version of the astSED routine that uses priors to throw out unphysical solutions

    Assuming BC03 here
    """

    modelFlux=[]
    modelMasses=[]
    for modelSEDDict in modelSEDDictList:
        modelFlux.append(modelSEDDict['flux'])
        modelMasses.append(modelSEDDict['stellarMass'])
    modelFlux=np.array(modelFlux)
    modelMasses=np.array(modelMasses)
    sedFlux=np.array([SEDDict['flux']]*len(modelFlux))
    sedFluxErr=np.array([SEDDict['fluxErr']]*len(modelFlux))

    # Analytic expression below is for normalisation at minimum chi squared (see note book)
    norm=np.sum((modelFlux*sedFlux)/(sedFluxErr**2), axis=1)/np.sum(modelFlux**2/sedFluxErr**2, axis=1)
    norms=np.array([norm]*modelFlux.shape[1]).transpose()
    chiSq=np.sum(((sedFlux-norms*modelFlux)**2)/sedFluxErr**2, axis=1)
    chiSq[np.isnan(chiSq)]=1e6   # throw these out, should check this out and handle more gracefully
    minChiSq=chiSq.min()
    bestMatchIndex=np.equal(chiSq, minChiSq).nonzero()[0][0]
    bestNorm=norm[bestMatchIndex]
    bestChiSq=minChiSq
    bestChiSqContrib=((sedFlux[bestMatchIndex]-norms[bestMatchIndex]*modelFlux[bestMatchIndex])**2)\
                        /sedFluxErr[bestMatchIndex]**2

    # Stellar mass
    stellarMasses=norm*distNorm*modelMasses
    bestStellarMass=stellarMasses[bestMatchIndex]

    fitResult={'minChiSq': bestChiSq,
                'chiSqContrib': bestChiSqContrib,
                'allChiSqValues': chiSq,
                'ageGyr': modelSEDDictList[bestMatchIndex]['ageGyr'],
                'norm': bestNorm,
                'log10StellarMass': np.log10(bestStellarMass),
                'z': modelSEDDictList[bestMatchIndex]['z'],
                'EBMinusV': modelSEDDictList[bestMatchIndex]['EBMinusV'],
                'tauGyr': modelSEDDictList[bestMatchIndex]['tauGyr'],
                'metallicity': modelSEDDictList[bestMatchIndex]['metallicity'],
                'bestFitModelSEDDict': modelSEDDictList[bestMatchIndex]}

    modelPath=modelSEDDictList[bestMatchIndex]['fileName']
    bestFitModel=astSED.BC03Model(modelPath)

    s=bestFitModel.getSED(fitResult['ageGyr'], z = fitResult['z'])
    s.extinctionCalzetti(fitResult['EBMinusV'])
    s.flux=s.flux*fitResult['norm']
    fitResult['bestFitSED']=s

    # Error bar by weighted chi-sq probability
    # NOTE: This seems to be the problem in crashing when multi-processing?
    #prob=stats.chisqprob(chiSq, 1)
    #logMasses=np.log10(stellarMasses)
    #massBinEdges=np.linspace(8.0, 13.5, 800)
    #massBinCentres=(massBinEdges[1]-massBinEdges[0])/2.0+massBinEdges[:-1]
    #binProb=np.zeros(len(massBinCentres))
    #for i in range(len(massBinCentres)):
        #binMin=massBinEdges[i]
        #binMax=massBinEdges[i+1]
        #mask=np.logical_and(np.greater(logMasses, binMin), np.less(logMasses, binMax))
        #binProb[i]=prob[mask].sum()
    #norm=np.trapz(binProb, massBinCentres)
    #binProb=binProb/norm
    #maxLikelihoodMass=massBinCentres[np.where(binProb == binProb.max())]
    #myProb=0.0
    #count=0
    #peakIndex=np.where(binProb == binProb.max())[0][0]
    #while myProb < 0.68:
        #count=count+1
        #myProb=np.trapz(binProb[peakIndex-count:peakIndex+count], massBinCentres[peakIndex-count:peakIndex+count])
    #minMass=massBinCentres[peakIndex-count]
    #maxMass=massBinCentres[peakIndex+count]
    #errLogStellarMass=np.max([maxMass-maxLikelihoodMass, maxLikelihoodMass-minMass])
    #fitResult['errLog10StellarMass_fromChiSq']=errLogStellarMass

    # If commented out above
    fitResult['errLog10StellarMass_fromChiSq']=-99

    return fitResult

#-------------------------------------------------------------------------------------------------------------
def makeSEDPlot(SEDDict, fitResult, plotFileName, title = ""):
    """Make a SED plot with the best fit shown

    """

    # Convert error bars to +/- format, so can avoid going -ve so matplotlib can plot them
    s=fitResult['bestFitSED']
    xMin=3500
    xMax=18000
    label="Age = %.1f Gyr, Z/Z$_{\odot}$ = %.3f, $\\tau$ = %.1f, $EBMinusV$ = %.2f, log($M^*$/M$_{\odot}$) = %.2f, $\chi^2$ = %.1f" % (fitResult['ageGyr'], fitResult['metallicity'], fitResult['tauGyr'], fitResult['EBMinusV'], fitResult['log10StellarMass'], fitResult['minChiSq'])
    pylab.plot(s.wavelength, s.flux, 'k-', label = label)
    errPlusMinus=np.zeros([2, SEDDict['fluxErr'].shape[0]])
    errPlusMinus[0]=SEDDict['fluxErr']
    errPlusMinus[1]=SEDDict['fluxErr']
    errMask=np.less(SEDDict['flux']-SEDDict['fluxErr'], 0)
    errPlusMinus[0][errMask]=SEDDict['flux'][errMask]*0.9999
    pylab.errorbar(SEDDict['wavelength'], SEDDict['flux'], yerr = errPlusMinus,
                    fmt = "ro", mec = "k", ecolor = "k", linewidth = 1.5, mew = 1.5, ms = 8)
    pylab.xlim(xMin, xMax)
    pylab.xlabel("Wavelength ($\AA{}$)")
    pylab.ylabel("Flux Density (erg s$^{-1}$ cm$^{-2}$ $\AA^{-1}$)")
    pylab.legend(loc = 'upper left', prop = {'size': 9})
    rangeMask=np.logical_and(np.greater(s.wavelength, xMin), np.less(s.wavelength, xMax))
    yMin=s.flux[rangeMask].min()
    yMax=s.flux[rangeMask].max()*1.1
    pylab.ylim(yMin, yMax)
    pylab.title(title)
    pylab.savefig(plotFileName)
    pylab.close()

